from . import trainers
from .trainers import callbacks

from . import evaluation

import mlpug.mlpug_logging as logging

from mlpug.multi_processing import MultiProcessingManager
from mlpug.pytorch.xla.multi_processing import XLADistributedContext

if not isinstance(MultiProcessingManager.get_context(), XLADistributedContext):
    MultiProcessingManager.set_context(XLADistributedContext())



import os
import sys

import numpy as np

import torch
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as DDP
import torch.multiprocessing as mp

import torchvision as tv

from mlpug.examples.documentation.shared_args import base_argument_set, describe_args

# Import mlpug for Pytorch backend
import mlpug.pytorch as mlp

from basics.logging import get_logger


def load_data():
    transform = tv.transforms.ToTensor()

    training_data = tv.datasets.FashionMNIST('./mlpug-datasets-temp/',
                                             train=True,
                                             download=True,
                                             transform=transform)
    test_data = tv.datasets.FashionMNIST('./mlpug-datasets-temp/',
                                         train=False,
                                         download=True,
                                         transform=transform)

    return training_data, test_data


def create_callbacks_for(trainer,
                         experiment_name,
                         model_hyper_parameters,
                         is_primary,
                         validation_dataset,
                         progress_log_period):
    # At minimum you want to log the loss in the training progress
    # By default the batch loss and the moving average of the loss are calculated and logged
    loss_evaluator = mlp.evaluation.MetricEvaluator(trainer=trainer)
    callbacks = [
        mlp.callbacks.TrainingMetricsLogger(metric_evaluator=loss_evaluator),
        # Calculate validation loss only once per epoch over the whole dataset
        mlp.callbacks.TestMetricsLogger(validation_dataset,
                                        'validation',
                                        metric_evaluator=loss_evaluator,
                                        batch_level=False),
        mlp.callbacks.CheckpointManager(base_checkpoint_filename=experiment_name,
                                        batch_level=False,  # monitor per epoch
                                        metric_to_monitor="validation.dataset.loss",
                                        metric_monitor_period=1,  # every epoch
                                        create_checkpoint_every=0,  # We are only interested in the best model,
                                                                    # not the latest model
                                        archive_last_model_checkpoint_every=0,  # no archiving
                                        backup_before_override=False,
                                        model_hyper_parameters=model_hyper_parameters)
    ]

    # Only primary worker needs to log progress
    if is_primary:
        callbacks += [
            mlp.callbacks.LogProgress(log_period=progress_log_period, set_names=['training', 'validation']),
        ]

    return callbacks


def build_model(hidden_size=128):
    return torch.nn.Sequential(
        torch.nn.Flatten(),
        torch.nn.Linear(784, hidden_size),
        torch.nn.ReLU(),
        torch.nn.Linear(hidden_size, 10))


# MLPug needs a TrainModel that outputs the loss
class TrainModel(torch.nn.Module):
    def __init__(self, classifier, device):
        super(TrainModel, self).__init__()

        self.classifier = classifier
        self.device = device

        self.loss_func = torch.nn.CrossEntropyLoss()

    def forward(self, batch_data, evaluate_settings, inference_mode=None):
        images, true_labels = batch_data

        images = images.to(self.device)
        true_labels = true_labels.to(self.device)

        logits = self.classifier(images)
        return self.loss_func(logits, true_labels)


def worker_fn(rank, args, world_size):

    distributed = args.distributed
    is_primary = rank == 0

    mlp.logging.use_fancy_colors()
    
    # ########### EXPERIMENT SETUP ############
    torch.random.manual_seed(args.seed)  # For reproducibility

    if distributed:
        logger_name = f"[Device {rank}] {os.path.basename(__file__)}"
    else:
        logger_name = os.path.basename(__file__)

    logger = get_logger(logger_name)
    # ########################################

    # ############## DEVICE SETUP ##############
    use_cuda = torch.cuda.is_available()
    if use_cuda:
        torch.cuda.set_device(rank)

        if distributed:
            os.environ['MASTER_ADDR'] = 'localhost'
            os.environ['MASTER_PORT'] = '12355'

            dist.init_process_group(backend='nccl', rank=rank, world_size=world_size)

            logger.info(f"Training using multiple GPUs: Using GPU {rank}/{world_size}")
        else:
            logger.info(f"Single device mode : Using GPU {rank} ")
    else:
        if distributed:
            logger.error(f"No GPUs available for data distributed training over multiple GPUs")
            return

        logger.info(f"Single device mode : Using CPU")

    device = torch.device("cuda" if use_cuda else "cpu")
    # ########################################

    # ########## SETUP BATCH DATASETS ##########
    if distributed and not is_primary:
        dist.barrier()

    training_data, test_data = load_data()

    if distributed and is_primary:
        dist.barrier()

    training_sampler = None
    validation_sampler = None
    if distributed:
        training_sampler = torch.utils.data.distributed.DistributedSampler(training_data)
        validation_sampler = torch.utils.data.distributed.DistributedSampler(test_data)

    training_dataset = torch.utils.data.DataLoader(training_data,
                                                   batch_size=args.batch_size,
                                                   shuffle=(training_sampler is None),
                                                   sampler=training_sampler,
                                                   num_workers=3)

    # Using the test set as a validation set, just for demonstration purposes
    validation_dataset = torch.utils.data.DataLoader(test_data,
                                                     batch_size=args.batch_size,
                                                     shuffle=(validation_sampler is None),
                                                     sampler=validation_sampler,
                                                     num_workers=3)
    # ##########################################

    # ############ BUILD THE MODEL #############
    classifier = build_model(args.hidden_size)

    train_model = TrainModel(classifier, device)

    # Move model to assigned GPU (see torch.cuda.set_device(args.local_rank))
    classifier.to(device)
    if distributed:
        train_model = DDP(train_model, device_ids=[rank])
    # ############################################

    # ############ SETUP OPTIMIZER #############
    optimizer = torch.optim.Adam(classifier.parameters(), lr=args.learning_rate)
    # ##########################################

    # ############# SETUP TRAINING ##############
    trainer = mlp.trainers.DefaultTrainer(optimizers=optimizer, model_components=classifier)

    model_hyper_parameters = {
        "hidden_size": args.hidden_size
    }

    callbacks = create_callbacks_for(trainer,
                                     args.experiment_name,
                                     model_hyper_parameters,
                                     is_primary,
                                     validation_dataset,
                                     args.progress_log_period)

    manager = mlp.trainers.TrainingManager(trainer,
                                           training_dataset,
                                           num_epochs=args.num_epochs,
                                           callbacks=callbacks,
                                           experiment_data={
                                               "args": args
                                           })

    trainer.set_training_model(train_model)
    # ##########################################

    # ################# START! #################
    manager.start_training()
    # ##########################################

    logger.info("DONE.")


def test_model(model_checkpoint_filename, logger, device=None):
    if device is None:
        device = torch.device("cpu")

    logger.info(f'Loading model checkpoint ...')
    checkpoint = torch.load(model_checkpoint_filename, map_location=device)

    # Contains 'hidden_size'
    classifier = build_model(**checkpoint['hyper_parameters'])
    classifier.load_state_dict(checkpoint['model'])

    _, test_data = load_data()

    first_sample = next(iter(test_data))
    image = first_sample[0]
    real_label = first_sample[1]

    classifier.eval()
    with torch.no_grad():
        logits = classifier(image)
        probabilities = torch.softmax(logits, dim=-1)

        predicted_label = torch.argmax(probabilities)

        logger.info(f"real label = {real_label}, predicted label = {predicted_label}\n")


if __name__ == '__main__':
    # ############# SETUP LOGGING #############
    mlp.logging.use_fancy_colors()
    logger = get_logger(os.path.basename(__file__))
    # ########################################

    # ############## PARSE ARGS ##############
    parser = base_argument_set()

    parser.parse_args()

    args = parser.parse_args()

    describe_args(args, logger)

    # ############## TRAIN MODEL ##############
    if args.distributed:
        num_gpus_available = torch.cuda.device_count()
        world_size = args.num_devices if args.num_devices > 0 else num_gpus_available
        if world_size > num_gpus_available:
            logger.warn(f"Number of requested GPUs is lower than available GPUs, "
                        f"limiting training to {num_gpus_available} GPUS")
            world_size = num_gpus_available

        logger.info(f"Distributed Data Parallel mode : Using {world_size} GPUs")
        logger.info(f"Global batch size: {args.batch_size*world_size}")

        mp.spawn(worker_fn,
                 args=(args, world_size,),
                 nprocs=world_size,
                 join=True)
    else:
        logger.info(f"Single device mode.")
        logger.info(f"Global batch size: {args.batch_size}")

        worker_fn(0, args=args, world_size=1)

    # ######### USE THE TRAINED MODEL ##########
    sys.stdout.write("\n\n\n")
    sys.stdout.flush()

    logger.info("Using the trained classifier ...")

    model_checkpoint_filename = f'../trained-models/{args.experiment_name}-best-model-checkpoint.pt'

    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda" if use_cuda else "cpu")

    test_model(model_checkpoint_filename, device=device, logger=logger)



import torch
import torch.nn as nn

from mlpug.examples.chatbot.conversation_parsing import normalizeString
from mlpug.examples.chatbot.conversation_dataset import indexesFromSentence


class GreedySearchDecoder(nn.Module):
    def __init__(self, encoder, decoder, SOS_token, device, select_token_func=None):
        super(GreedySearchDecoder, self).__init__()
        self.encoder = encoder
        self.decoder = decoder

        self.SOS_token = SOS_token

        self.device = device

        self.select_token_func = select_token_func
        if not self.select_token_func:
            self.select_token_func = lambda output: torch.max(output, dim=2)

    def forward(self, input_seq, input_length, max_length):
        # Forward input through encoder model
        encoder_outputs, encoder_hidden = self.encoder(input_seq, input_length)
        # Prepare encoder's final hidden layer to be first hidden input to the decoder
        # decoder_hidden = encoder_hidden[:self.decoder.n_layers]
        decoder_n_layers = self.decoder.n_layers
        decoder_hidden = 0.5 * (encoder_hidden[:decoder_n_layers, 0, :, :] + encoder_hidden[:decoder_n_layers, 1, :, :])

        # Initialize decoder input with SOS_token
        decoder_input = torch.ones(1, 1, device=self.device, dtype=torch.long) * self.SOS_token
        # Initialize tensors to append decoded words to
        all_tokens = torch.zeros([0], device=self.device, dtype=torch.long)
        all_scores = torch.zeros([0], device=self.device)
        # Iteratively decode one word token at a time
        for _ in range(max_length):
            # Forward pass through decoder
            decoder_output, decoder_hidden = self.decoder(decoder_input, decoder_hidden, encoder_outputs)
            # Obtain most likely word token and its softmax score
            decoder_scores, decoder_input = self.select_token_func(decoder_output)

            # Record token and score
            all_tokens = torch.cat((all_tokens, decoder_input), dim=0)
            all_scores = torch.cat((all_scores, decoder_scores), dim=0)

        # Return collections of word tokens and scores
        return all_tokens, all_scores


def evaluate(encoder, decoder, searcher, voc, sentence, max_length, EOS_token, device):
    ### Format input sentence as a batch
    # words -> indexes
    indexes_batch = [indexesFromSentence(voc, sentence, EOS_token)]
    # Create lengths tensor
    lengths = torch.tensor([len(indexes) for indexes in indexes_batch])
    # Batch dimension should be second (in order for portioning over multiple GPUs)
    # TODO : that means unsqueeze(1)
    lengths = lengths.unsqueeze(0)

    # Transpose dimensions of batch to match models' expectations
    input_batch = torch.LongTensor(indexes_batch).transpose(0, 1)
    # Use appropriate device
    input_batch = input_batch.to(device)
    lengths = lengths.to(device)
    # Decode sentence with searcher
    tokens, scores = searcher(input_batch, lengths, max_length)
    # indexes -> words
    decoded_words = [voc.index2word[token.item()] for token in tokens]
    return decoded_words


def evaluateInput(encoder, decoder, searcher, voc, max_length, EOS_token, device):
    input_sentence = ''
    while True:
        try:
            # Get input sentence
            input_sentence = input('> ')
            # Check if it is quit case
            if input_sentence == '*q*' or input_sentence == '*quit*':
                break
            # Normalize sentence
            input_sentence = normalizeString(input_sentence)
            # Evaluate sentence
            output_words = evaluate(encoder, decoder, searcher, voc, input_sentence, max_length, EOS_token, device)

            # output_words[:] = [x for x in output_words if not (x == 'EOS' or x == 'PAD')]

            # Format and print response sentence
            response = []
            for word in output_words:
                if word == 'EOS':
                    break
                response.append(word)

            print('Bot:', ' '.join(response))

        except KeyError:
            print("Error: Encountered unknown word.")



import os

import abc

import torch
import torch.distributed as dist

from mlpug.trainers.training import BatchChunkingResults
from mlpug.evaluation import default_metric_reducer_func, MetricEvaluatorBase

from mlpug.utils import is_chunkable

from mlpug.base import Base

from mlpug.pytorch.multi_processing import MultiProcessingMixin


# ####### DEFAULT GATHER LOSS METHODS ########
class GatherLossBase(MultiProcessingMixin, Base, metaclass=abc.ABCMeta):

    def __init__(self,
                 name,
                 delete_training_loss=True,
                 delete_auxiliary_results=True,
                 requester=None,
                 **kwargs):

        if requester is not None:
            name += f'[{requester}]'

        super(GatherLossBase, self).__init__(pybase_logger_name=name, **kwargs)

        self._delete_training_loss = delete_training_loss
        self._delete_auxiliary_results = delete_auxiliary_results

        self.requester = requester

        self._gather_loss_func = None
        if self.is_distributed:
            self._log.info(f"Using distributed gather loss function")
            self._gather_loss_func = self._gather_loss_distributed
        else:
            self._log.info(f"Using gather loss function")
            self._gather_loss_func = self._gather_loss

    def _do_detatch_auxiliary_results(self, auxiliary_results):
        # When auxiliary_results is a BatchChunkingResults list, it was created by batch chunking
        if type(auxiliary_results) is BatchChunkingResults:
            for aux in auxiliary_results:
                aux[0].detach_()
                aux[1].detach_()
        else:
            auxiliary_results[0].detach_()
            auxiliary_results[1].detach_()

    def _do_delete_auxiliary_results(self, auxiliary_results):
        # When auxiliary_results is a BatchChunkingResults list, it was created by batch chunking
        if type(auxiliary_results) is BatchChunkingResults:
            for (ls, ns) in auxiliary_results:
                del ls
                del ns
        else:
            (ls, ns) = auxiliary_results
            del ls
            del ns

    @abc.abstractmethod
    def __call__(self, loss, auxiliary_results, **kwargs):
        raise NotImplementedError()

    @abc.abstractmethod
    def _gather_loss(self, **kwargs):
        raise NotImplementedError()

    @abc.abstractmethod
    def _gather_loss_distributed(self, **kwargs):
        raise NotImplementedError()


class GatherLossSimple(GatherLossBase):

    def __init__(self, delete_training_loss=True, requester=None, name="GatherLossSimple", **kwargs):
        super().__init__(name,
                         delete_training_loss=delete_training_loss,
                         requester=requester,
                         **kwargs)

    def __call__(self, loss, **kwargs):
        training_loss = loss
        training_loss.detach_()

        loss, loss_sum, num_samples = self._gather_loss_func(loss=training_loss, **kwargs)

        if self._delete_training_loss:
            del training_loss

        return loss, loss_sum, num_samples

    def _gather_loss(self, loss, **kwargs):
        loss = loss.item()
        return loss, loss, 1

    def _gather_loss_distributed(self, loss, **kwargs):
        loss_sum = loss
        dist.all_reduce(loss_sum)
        num_devices = dist.get_world_size()
        loss = loss_sum / num_devices

        return loss.item(), loss_sum.item(), num_devices


class GatherMaskedLoss(GatherLossBase):

    def __init__(self,
                 delete_training_loss=True,
                 delete_auxiliary_results=True,
                 requester=None,
                 name="GatherMaskedLoss",
                 **kwargs):
        super().__init__(name, delete_training_loss, delete_auxiliary_results, requester, **kwargs)

    def __call__(self, loss, auxiliary_results, **kwargs):
        training_loss = loss
        training_loss.detach_()

        self._do_detatch_auxiliary_results(auxiliary_results)

        # When auxiliary_results is a BatchChunkingResults list, it was created by batch chunking
        if type(auxiliary_results) is BatchChunkingResults:
            loss_sum = 0
            num_samples = 0
            for aux in auxiliary_results:
                loss_sum += aux[0]
                num_samples += aux[1]
        else:
            loss_sum = auxiliary_results[0]
            num_samples = auxiliary_results[1]

        loss, loss_sum, num_samples = self._gather_loss_func(loss_sum=loss_sum, num_samples=num_samples, **kwargs)

        if self._delete_training_loss:
            del training_loss

        if self._delete_auxiliary_results:
            self._do_delete_auxiliary_results(auxiliary_results)

        return loss, loss_sum, num_samples

    def _gather_loss(self, loss_sum, num_samples, **kwargs):
        loss_sum = loss_sum.item()
        num_samples = num_samples.item()

        loss = loss_sum/num_samples

        return loss, loss_sum, num_samples

    def _gather_loss_distributed(self, loss_sum, num_samples, **kwargs):
        dist.all_reduce(loss_sum)
        dist.all_reduce(num_samples)

        loss_sum = loss_sum.item()
        num_samples = num_samples.item()

        loss = loss_sum / num_samples

        return loss, loss_sum, num_samples
# ############################################


class MetricEvaluator(MultiProcessingMixin, MetricEvaluatorBase):

    def __init__(self, *args, batch_metric_funcs=None, name="MetricEvaluator", **kwargs):

        if batch_metric_funcs is None:
            batch_metric_funcs = {
                "loss": GatherLossSimple(requester=name)
            }

        super().__init__(batch_metric_funcs, *args, name=name, **kwargs)

    def _create_default_model_evaluate_func(self):

        def evaluate_loss(batch, evaluate_settings=None):
            if is_chunkable(batch):
                # Get raw batch
                batch = batch[:]

            with torch.no_grad():
                results = self._trainer.evaluate_loss(
                    batch,
                    inference_mode=True,
                    evaluate_settings=evaluate_settings)

            return results

        return evaluate_loss




import os
import random

import numpy as np

import torch
import torch.nn as nn
from torch import optim
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as PT_DDP

from mlpug.examples.chatbot.shared import create_argument_parser

from mlpug.examples.chatbot.pytorch.original_chatbot_tutorial.model_data_generation import \
    create_sentence_pairs_collate_fn
from mlpug.examples.chatbot.conversation_dataset import IndexedSentencePairsDataset, load_sentence_pair_data

from mlpug.examples.chatbot.pytorch.original_chatbot_tutorial.seq2seq import EncoderRNN, LuongAttnDecoderRNN
from mlpug.examples.chatbot.pytorch.original_chatbot_tutorial.training import Seq2SeqTrainModel
from mlpug.examples.chatbot.pytorch.original_chatbot_tutorial.loss import masked_loss

import mlpug.pytorch as mlp

from mlpug.utils import get_value_at

from basics.logging import get_logger


def transfer_to_gpu(opt):
    # If you have cuda, configure cuda to call
    for state in opt.state.values():
        for k, v in state.items():
            if isinstance(v, torch.Tensor):
                state[k] = v.cuda()


def dataset_path_for(subset):
    return os.path.join(dataset_path, f"{subset}-{base_dataset_filename}")


class TeacherForcingController(mlp.callbacks.Callback):

    def on_epoch_start(self, logs):
        if 'training_settings' not in logs:
            logs['training_settings'] = {}
            logs['evaluate_settings'] = {}

        return True

    def on_batch_training_start(self, training_batch, logs):
        # Determine if we are using teacher forcing this iteration
        use_teacher_forcing = random.random() < teacher_forcing_ratio

        logs['training_settings']['use_teacher_forcing'] = use_teacher_forcing
        logs['evaluate_settings']['use_teacher_forcing'] = use_teacher_forcing

        return True


class ChatbotTrainer(mlp.trainers.DefaultTrainer):

    def _evaluate_loss(self, batch_data, evaluate_settings=None, inference_mode=None):

        use_teacher_forcing = get_value_at('use_teacher_forcing', evaluate_settings)
        if use_teacher_forcing is None:
            use_teacher_forcing = True

        padded_input_batch, input_lengths, output_batch, output_mask, max_output_len = batch_data

        padded_input_batch = padded_input_batch.to(device)
        input_lengths = input_lengths.to(device)
        output_batch = output_batch.to(device)
        output_mask = output_mask.to(device)

        batch_size = padded_input_batch.size(1)

        init_decoder_input = torch.tensor([[SOS_token for _ in range(batch_size)]], dtype=torch.long, device=device)

        results = self.training_model(padded_input_batch, input_lengths, init_decoder_input, max_output_len,
                                      output_batch, use_teacher_forcing)

        per_sample_loss = results["loss"]

        # loss, loss_sum, num_samples
        return masked_loss(per_sample_loss, output_mask)

    def _prepare_update_model_parameters(self):
        if not willClip:
            return

        # Clip gradients: gradients are modified in place
        if not use_mixed_precision:
            _ = nn.utils.clip_grad_norm_(self.training_model.parameters(), clip)
        else:
            for optimizer in self.optimizers.values():
                # Unscales the gradients of optimizer's assigned params in-place
                self._scaler.unscale_(optimizer)

                # Since the gradients of optimizer's assigned params are unscaled, clips as usual:
                nn.utils.clip_grad_norm_(self.training_model.parameters(), clip)


if __name__ == "__main__":

    parser = create_argument_parser()

    parser.add_argument(
        '--attention-model',
        type=str, required=False, default='general',
        help='Attention model type: "dot", "general" or "concat"')

    parser.add_argument(
        '--local_rank',
        type=int, required=False, default=None,
        help='Set automatically when running this script in distributed parallel mode')

    parser.add_argument(
        '--teacher-forcing-ratio',
        type=float, required=False, default=1.0,
        help='How often teacher forcing should be applied, compared to not applying it')

    args = parser.parse_args()

    num_gpus = args.num_gpus
    use_mixed_precision = args.float16
    distributed = args.local_rank is not None
    single_process_data_parallel = (num_gpus is not None) and (num_gpus > 1)
    use_cuda = torch.cuda.is_available()

    is_first_worker = (distributed and args.local_rank == 0) or (not distributed)

    logger_name = os.path.basename(__file__)
    if distributed:
        logger_name += f" (WORKER {args.local_rank})"

    logger = get_logger(logger_name)
    mlp.logging.use_fancy_colors()

    if distributed and single_process_data_parallel:
        logger.warn("The num_gpus argument will be ignored in Distributed Data Parallel mode. "
                    "Number of GPUs will be defined through --nproc_per_node argument of the "
                    "torch.distributed.launch script")
        num_gpus = None
        single_process_data_parallel = False

    if distributed and not use_cuda:
        logger.error("Distributed Data Parallel mode can only be used when NVidia GPUs are available")
        exit(-1)

    if use_mixed_precision and single_process_data_parallel:
        logger.error("Mixed precision training can't be combined with single process data parallel, "
                     "use distributed data parallel mode instead")
        exit(-1)

    if args.remote_debug and is_first_worker:
        import pydevd_pycharm
        pydevd_pycharm.settrace('192.168.178.85', port=57491, stdoutToServer=True, stderrToServer=True)

    seed = args.seed
    logger.info(f"Seed : {seed}")
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)

    logger.info(f"Using CUDA? {use_cuda}")
    logger.info(f"In Distributed Data Parallel mode ? {distributed}")
    logger.info(f"Using single process Data Parallel mode "
                f"(multiple GPUs in one process)? {single_process_data_parallel}\n\n")
    logger.info(f"Use mixed precision training? {use_mixed_precision}")

    if distributed:
        logger.info(f"Distributed Data Parallel mode : Using GPU {args.local_rank} ")
        torch.cuda.set_device(args.local_rank)

        dist.init_process_group(backend='nccl', init_method='env://')

    device = torch.device("cuda" if use_cuda else "cpu")

    ##################################################
    #
    # [START] Setup
    #
    ##################################################

    # ############ Conversations dataset #############
    # Default word tokens

    PAD_token = 0  # Used for padding short sentences
    SOS_token = 1  # Start-of-sentence token
    EOS_token = 2  # End-of-sentence token

    dataset_path = args.dataset_path
    base_dataset_filename = args.base_dataset_filename
    logger.info(f"dataset_path : {dataset_path}")
    logger.info(f"base_dataset_filename : {base_dataset_filename}")

    ##################################################

    # ############ Model configuration ###############
    embedding_size = args.embedding_size
    logger.info(f"embedding_size : {embedding_size}")

    state_size = args.state_size
    logger.info(f"state_size : {state_size}")

    num_layers = args.num_layers
    logger.info(f"num_layers : {num_layers}")

    attn_model = args.attention_model
    logger.info(f"Attention model : {attn_model}")

    dropout = args.dropout
    logger.info(f"dropout rate : {dropout}")
    ##################################################

    # ########### Training/optimization ##############
    experiment_name = args.experiment_name
    logger.info(f"experiment_name: {experiment_name}")

    batch_size = args.batch_size
    if single_process_data_parallel:
        batch_size = num_gpus * batch_size
        logger.info(f"batch_size (all GPUs combined): {batch_size}")
    else:
        logger.info(f"batch_size (per process): {batch_size}")

    clip = args.gradient_clipping
    willClip = clip > 0.0
    logger.info(f"Gradient clipping : {clip} (Will clip? {willClip})")

    teacher_forcing_ratio = args.teacher_forcing_ratio
    logger.info(f"Teacher forcing ratio : {teacher_forcing_ratio}")

    learning_rate = args.learning_rate
    logger.info(f"Base learning rate : {learning_rate}")

    decoder_learning_ratio = args.decoder_learning_rate_ratio
    logger.info(f"Decoder LR ratio : {learning_rate}")

    num_epochs = args.num_epochs
    logger.info(f"Number of training epochs : {num_epochs}")

    progress_logging_period = args.progress_logging_period
    logger.info(f"Progress logging period : {progress_logging_period}")

    # For Tensorboard
    metric_names = {
        'batch.loss': 'cross_entropy',
        'duration.batch': 'training_time',
        'duration.window_average': 'training_time',
        'duration.dataset': 'training_time',
        'batch_size': 'size'
    }

    ##################################################

    ##################################################
    #
    # [END] Setup
    #
    ##################################################

    # ############## SETUP DATASETS ###################
    logger.info('Setup data sets ...')

    logger.info('Loading training set ...')
    training_dataset, voc = load_sentence_pair_data(dataset_path_for('training'), logger)
    logger.info('Loading validation set ...')
    validation_dataset, _unused_ = load_sentence_pair_data(dataset_path_for('validation'), logger)

    logger.debug(f"Number of sentence pairs in training set: {len(training_dataset)}")
    logger.debug(f"Number of sentence pairs in validation set: {len(validation_dataset)}")

    training_dataset = IndexedSentencePairsDataset(training_dataset, voc, EOS_token)
    validation_dataset = IndexedSentencePairsDataset(validation_dataset, voc, EOS_token)

    # collate_sentence_pairs = create_sentence_pairs_collate_fn(PAD_token, fixed_sequence_length=40)
    collate_sentence_pairs = create_sentence_pairs_collate_fn(PAD_token)

    training_sampler = None
    validation_sampler = None
    if distributed:
        training_sampler = torch.utils.data.distributed.DistributedSampler(training_dataset)
        validation_sampler = torch.utils.data.distributed.DistributedSampler(validation_dataset)

    train_dataset_loader = torch.utils.data.DataLoader(training_dataset,
                                                       batch_size=batch_size,
                                                       shuffle=(training_sampler is None),
                                                       num_workers=3,
                                                       collate_fn=collate_sentence_pairs,
                                                       pin_memory=True,
                                                       sampler=training_sampler)

    validation_dataset_loader = torch.utils.data.DataLoader(validation_dataset,
                                                            batch_size=batch_size,
                                                            shuffle=(validation_sampler is None),
                                                            num_workers=3,
                                                            collate_fn=collate_sentence_pairs,
                                                            pin_memory=True,
                                                            sampler=validation_sampler)

    ##################################################

    # ############### BUILD MODEL ####################
    logger.info('Building model ...')

    # Initialize word embeddings
    embedding = nn.Embedding(voc.num_words, embedding_size)

    # Initialize encoder & decoder models
    encoder = EncoderRNN(state_size, embedding, num_layers, dropout)
    decoder = LuongAttnDecoderRNN(attn_model, embedding, state_size, voc.num_words, num_layers, dropout)

    training_model = Seq2SeqTrainModel(encoder, decoder)

    logger.info(f"Transfer training model to device {device} ...")
    # Use appropriate device
    training_model = training_model.to(device)

    ##################################################

    # ############## SETUP TRAINING ##################

    # Initialize optimizers
    logger.info('Building optimizers ...')
    encoder_optimizer = optim.Adam(encoder.parameters(), lr=learning_rate)
    decoder_optimizer = optim.Adam(decoder.parameters(), lr=learning_rate * decoder_learning_ratio)

    # TODO : Is this required for DDP and mixed procession training?
    if use_cuda:
        logger.info(f"Transfer optimizers model to {device} ...")
        transfer_to_gpu(encoder_optimizer)
        transfer_to_gpu(decoder_optimizer)

    if single_process_data_parallel:
        logger.info(f"We are using {num_gpus} GPUs in a single process, making training model data parallel ...")
        # TODO : actually use the right number of GPUs
        training_model = nn.DataParallel(training_model, dim=1)
    elif distributed:
        training_model = PT_DDP(training_model,
                                device_ids=[args.local_rank],
                                output_device=args.local_rank,
                                dim=1)

    logger.info('Prepare training ...')

    trainer = ChatbotTrainer([encoder_optimizer, decoder_optimizer],
                             {
                                 'embedding': embedding,
                                 'encoder': encoder,
                                 'decoder': decoder
                             },
                             use_mixed_precision=use_mixed_precision)

    average_loss_evaluator = mlp.evaluation.MetricEvaluator(trainer=trainer,
                                                            batch_metric_funcs={
                                                                "loss": mlp.evaluation.GatherMaskedLoss()
                                                            },
                                                            name="AverageLossEvaluator")

    callbacks = []
    if distributed:
        callbacks += [mlp.callbacks.DistributedSamplerManager(training_sampler, name="SamplerManager[training]"),
                      mlp.callbacks.DistributedSamplerManager(validation_sampler, name="SamplerManager[validation]")]

    callbacks += [TeacherForcingController(name="TeacherForcingController"),
                  mlp.callbacks.TrainingMetricsLogger(metric_evaluator=average_loss_evaluator),
                  mlp.callbacks.TestMetricsLogger(validation_dataset_loader,
                                                  dataset_name='validation',
                                                  metric_evaluator=average_loss_evaluator)]

    if is_first_worker:
        callbacks += [
            mlp.callbacks.BatchSizeLogger(),
            mlp.callbacks.CheckpointManager(metric_to_monitor='validation.window_average.loss',
                                            base_checkpoint_filename=args.experiment_name,
                                            archive_last_model_checkpoint_every=20000),
            mlp.callbacks.LogProgress(log_period=progress_logging_period, set_names=["training", "validation"]),
            mlp.callbacks.AutoTensorboard(experiment_name=experiment_name, dataset_name='training',
                                          metric_names=metric_names),
            mlp.callbacks.AutoTensorboard(experiment_name=experiment_name, dataset_name='validation',
                                          metric_names=metric_names),
            # Batch-level batch duration and batch size
            mlp.callbacks.Tensorboard(['duration.batch', 'batch_size'],
                                      experiment_name=experiment_name,
                                      dataset_name='training_params',
                                      metric_names=metric_names,
                                      ignore_missing_metrics=True),
            # Batch-level average batch duration
            mlp.callbacks.Tensorboard(['duration.window_average'],
                                      experiment_name=experiment_name,
                                      dataset_name='training_params',
                                      metrics_are_averages=True,
                                      metric_names=metric_names,
                                      ignore_missing_metrics=True),
            # Epoch-level epoch duration
            mlp.callbacks.Tensorboard(['duration.dataset'],
                                      experiment_name=experiment_name,
                                      dataset_name='training_params',
                                      batch_level=False,
                                      metric_names=metric_names,
                                      ignore_missing_metrics=True)]

    manager = mlp.trainers.TrainingManager(trainer,
                                           train_dataset_loader,
                                           num_epochs=num_epochs,
                                           callbacks=callbacks,
                                           experiment_data=args)

    tc_file = args.training_checkpoint
    if tc_file:
        # Can we make this whole procedure a part of MLPug?
        logger.info(f"Loading training checkpoint : {tc_file}")
        map_location = None
        if distributed:
            map_location = f'cuda:{args.local_rank}'

        checkpoint = torch.load(tc_file, map_location=map_location)

        manager.set_state(checkpoint)

        if distributed:
            dist.barrier()
        logger.info(f"Ready loading training checkpoint.")

        # Required, else loading a checkpoint can lead to OOMs
        del checkpoint
        if use_cuda:
            torch.cuda.empty_cache()

    trainer.set_training_model(training_model)

    ##################################################

    logger.info('Start training ...')
    manager.start_training()



import argparse


def create_argument_parser(description='Train reference chatbot on processed sentence pair data'):
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        '--experiment-name',
        type=str, required=True,
        help='Name of experiment, used as file name base for checkpoints')

    parser.add_argument(
        '--dataset-path',
        type=str, required=True,
        help='Path to dataset files')

    parser.add_argument(
        '--base-dataset-filename',
        type=str, required=True,
        help='Shared filename postfix for training and validation set')

    parser.add_argument(
        '--training-checkpoint',
        type=str, required=False, default=None,
        help='Path to training checkpoint file')

    parser.add_argument('--float16',
                        action='store_true',
                        required=False,
                        help='Provide this flag to use Mixed Precision Training')

    parser.add_argument(
        '--num-gpus',
        type=int, required=False, default=None,
        help='The number of GPUs to use in one process (DataParallel will be used)')

    parser.add_argument(
        '--embedding-size',
        type=int, required=False, default=512,
        help='Input embedding size')

    parser.add_argument(
        '--state-size',
        type=int, required=False, default=768,
        help='Model state size')

    parser.add_argument(
        '--num-layers',
        type=int, required=False, default=2,
        help='Number of Encoder and Decoder layers')

    parser.add_argument(
        '--dropout',
        type=float, required=False, default=0.1,
        help='Drop out rate')

    parser.add_argument(
        '--batch-size',
        type=int, required=False, default=64,
        help='Batch size')

    parser.add_argument(
        '--gradient-clipping',
        type=float, required=False, default=50.0,
        help='Gradient clipping threshold')

    parser.add_argument(
        '--learning-rate',
        type=float, required=False, default=1e-4,
        help='Base learning rate')

    parser.add_argument(
        '--decoder-learning-rate-ratio',
        type=float, required=False, default=2.0,
        help='Decoder LR ratio, such that the decoder LR is ratio * base LR')

    parser.add_argument(
        '--num-epochs',
        type=int, required=False, default=20,
        help='Number of epochs to train')

    parser.add_argument(
        '--progress-logging-period',
        type=int, required=False, default=20,
        help='Period in global (batch) iterations before we log the training progress again.')

    parser.add_argument(
        '--seed',
        type=int, required=False, default=0,
        help='Random seed to use to ensure same random split at each restart')

    parser.add_argument(
        '--remote-debug',
        action='store_true',
        help="If specified, remote debugging will be enabled")

    return parser



from mlpug.pytorch.trainers.callbacks.checkpoint_manager import CheckpointManager as CheckpointManagerPyTorch

import torch_xla.core.xla_model as xm


class CheckpointManager(CheckpointManagerPyTorch):

    def __init__(self,
                 *args,
                 is_primary=None,
                 **kwargs):

        if is_primary is None:
            is_primary = xm.is_master_ordinal()

        super().__init__(*args, is_primary=is_primary, **kwargs)

    def _save_model_checkpoint(self, filename, state):
        # By default only the master worker saves
        xm.save(state, filename)

    def _save_training_checkpoint(self, filename, state):
        # By default only the master worker saves
        xm.save(state, filename)



from mlpug.pytorch.multi_processing import MultiProcessingMixin

from mlpug.trainers.callbacks.callback import *
from mlpug.trainers.callbacks.callback import Callback as CallbackBase


class Callback(MultiProcessingMixin, CallbackBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import tensorflow as tf


class TrainModel(tf.keras.Model):
    def __init__(self, transformer):
        super(TrainModel, self).__init__()

        self.transformer = transformer

    def call(self, batch_data, evaluate_settings, inference_mode=None):
        inp, tar = batch_data

        tar_inp = tar[:, :-1]
        tar_real = tar[:, 1:]

        enc_padding_mask, combined_mask, dec_padding_mask = create_masks(inp, tar_inp)

        predictions, _ = self.transformer(inp, tar_inp, True, enc_padding_mask, combined_mask, dec_padding_mask)
        loss, summed_loss, num_samples = loss_function(tar_real, predictions)

        return loss, summed_loss, num_samples


def create_padding_mask(seq):
    seq = tf.cast(tf.math.equal(seq, 0), tf.float32)

    # add extra dimensions to add the padding
    # to the attention logits.
    return seq[:, tf.newaxis, tf.newaxis, :]  # (batch_size, 1, 1, seq_len)


def create_look_ahead_mask(size):
    mask = 1 - tf.linalg.band_part(tf.ones((size, size)), -1, 0)
    return mask  # (seq_len, seq_len)


def create_masks(inp, tar):
    # Encoder padding mask
    enc_padding_mask = create_padding_mask(inp)

    # Used in the 2nd attention block in the decoder.
    # This padding mask is used to mask the encoder outputs.
    dec_padding_mask = create_padding_mask(inp)

    # Used in the 1st attention block in the decoder.
    # It is used to pad and mask future tokens in the input received by
    # the decoder.
    look_ahead_mask = create_look_ahead_mask(tf.shape(tar)[1])
    dec_target_padding_mask = create_padding_mask(tar)
    combined_mask = tf.maximum(dec_target_padding_mask, look_ahead_mask)

    return enc_padding_mask, combined_mask, dec_padding_mask


loss_object = tf.keras.losses.SparseCategoricalCrossentropy(
    from_logits=True, reduction='none')


def loss_function(real, pred):
    mask = tf.math.logical_not(tf.math.equal(real, 0))
    loss_ = loss_object(real, pred)

    mask = tf.cast(mask, dtype=loss_.dtype)
    loss_ *= mask

    summed_loss = tf.reduce_sum(loss_)
    num_samples = tf.reduce_sum(mask)

    return summed_loss/num_samples, summed_loss, num_samples


class CustomSchedule(tf.keras.optimizers.schedules.LearningRateSchedule):
    def __init__(self, d_model, warmup_steps=4000):
        super(CustomSchedule, self).__init__()

        self.d_model = d_model
        self.d_model = tf.cast(self.d_model, tf.float32)

        self.warmup_steps = warmup_steps

    def __call__(self, step):
        arg1 = tf.math.rsqrt(step)
        arg2 = step * (self.warmup_steps ** -1.5)

        return tf.math.rsqrt(self.d_model) * tf.math.minimum(arg1, arg2)



from mlpug.pytorch.multi_processing import MultiProcessingMixin

from mlpug.mlpug_exceptions import CallbackBadUseException

from mlpug.trainers.callbacks.tensorboard import *
from mlpug.trainers.callbacks.tensorboard import \
    Tensorboard as TensorboardBase, \
    AutoTensorboard as AutoTensorboardBase


class Tensorboard(MultiProcessingMixin, TensorboardBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not self.is_primary:
            raise CallbackBadUseException(self.name, "Only add Tensorboard logging to the primary worker")


class AutoTensorboard(MultiProcessingMixin, AutoTensorboardBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not self.is_primary:
            raise CallbackBadUseException(self.name, "Only add AutoTensorboard logging to the primary worker")



import logging

from basics.logging import LOGGING_FORMAT

from mlpug.mlpug_logging.base import FG_COLOR_CODE, BG_COLOR_CODE, FormatterBase

WHITE = 15
PINK = 13
MINT = 14
GRAY = 248


class FancyColorFormatter(FormatterBase):
    LEVEL_BG_COLOR_MAP = {
        'WARNING': MINT,
        'ERROR': PINK
    }

    LEVEL_TEXT_COLOR_MAP = {
      'WARNING': WHITE,
      'DEBUG': GRAY,
      'CRITICAL': WHITE,
      'ERROR': WHITE
    }

    MSG_COLOR_MAP = {
      'DEBUG': GRAY
    }

    def format(self, record):
        level = record.levelname

        # Align length of level string
        level_aligned = level + ' '*max(8-len(level), 0)

        level_fmt = self._set_bg_color(level, level_aligned, self.LEVEL_BG_COLOR_MAP)
        level_fmt = self._set_fg_color(level, level_fmt, self.LEVEL_TEXT_COLOR_MAP)
        if level_aligned != level_fmt:
            level_fmt = self.finalize_fmt(level_fmt)

        record.levelname = level_fmt

        name = record.name
        name_fmt = self._set_fg_color(level, name, self.MSG_COLOR_MAP)
        if name != name_fmt:
            name_fmt = self.finalize_fmt(name_fmt)

        record.name = name_fmt

        func_name = record.funcName
        func_name_fmt = self._set_fg_color(level, func_name, self.MSG_COLOR_MAP)
        if func_name != func_name_fmt:
            func_name_fmt = self.finalize_fmt(func_name_fmt)

        record.funcName = func_name_fmt

        msg = record.msg
        msg_fmt = self._set_fg_color(level, msg, self.MSG_COLOR_MAP)
        if msg != msg_fmt:
            msg_fmt = self.finalize_fmt(msg_fmt)

        record.msg = msg_fmt

        return super().format(record)

    def _set_bg_color(self, level, s, color_map):
        return self.apply_fmt(BG_COLOR_CODE, level, s, color_map)

    def _set_fg_color(self, level, s, color_map):
        return self.apply_fmt(FG_COLOR_CODE, level, s, color_map)


def use_fancy_colors(log_level=logging.DEBUG):
    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)

    try:
        handler = root_logger.handlers[0]
        handler.setFormatter(FancyColorFormatter(LOGGING_FORMAT))
    except IndexError:
        print("ERROR : Unable to use fancy colors for logging, use logging.basicConfig or logging.config first")



# Partially based on https://github.com/tensorflow/docs/blob/master/site/en/tutorials/keras/classification.ipynb and
# https://colab.research.google.com/github/tensorflow/docs/blob/master/site/en/tutorials/distribute/custom_training.ipynb

import os
import sys

import pickle

import tensorflow as tf
import h5py
from tensorflow.python.keras.saving import hdf5_format

from basics.logging import get_logger

# Import mlpug for Tensorflow backend
import mlpug.tensorflow as mlp

from mlpug.examples.documentation.shared_args import base_argument_set, describe_args


def load_data():
    fashion_mnist = tf.keras.datasets.fashion_mnist
    (train_images, train_labels), (test_images, test_labels) = fashion_mnist.load_data()

    # Adding a dimension to the array -> new shape == (28, 28, 1)
    # We are doing this because the first layer in our model is a convolutional
    # layer and it requires a 4D input (batch_size, height, width, channels).
    # batch_size dimension will be added later on.
    train_images = train_images[..., None]
    test_images = test_images[..., None]

    # Scale
    train_images = train_images / 255.0
    test_images = test_images / 255.0

    return train_images, train_labels, test_images, test_labels


def create_callbacks_for(trainer,
                         experiment_name,
                         model_hyper_parameters,
                         distribution_strategy,
                         validation_dataset,
                         progress_log_period):
    # At minimum you want to log the loss in the training progress
    # By default the batch loss and the moving average of the loss are calculated and logged
    loss_evaluator = mlp.evaluation.MetricEvaluator(trainer=trainer,
                                                    distribution_strategy=distribution_strategy)
    callbacks = [
        mlp.callbacks.TrainingMetricsLogger(metric_evaluator=loss_evaluator),
        # Calculate validation loss only once per epoch over the whole dataset
        mlp.callbacks.TestMetricsLogger(validation_dataset,
                                        'validation',
                                        metric_evaluator=loss_evaluator,
                                        batch_level=False),
        mlp.callbacks.CheckpointManager(base_checkpoint_filename=experiment_name,
                                        batch_level=False,  # monitor per epoch
                                        metric_to_monitor="validation.dataset.loss",
                                        metric_monitor_period=1,  # every epoch
                                        create_checkpoint_every=0,  # We are only interested in the best model,
                                        # not the latest model
                                        archive_last_model_checkpoint_every=0,  # no archiving
                                        backup_before_override=False,
                                        model_hyper_parameters=model_hyper_parameters),
        mlp.callbacks.LogProgress(log_period=progress_log_period, set_names=['training', 'validation']),
    ]

    return callbacks


def build_model(hidden_size=128):
    return tf.keras.Sequential([
            tf.keras.layers.Flatten(input_shape=(28, 28)),
            tf.keras.layers.Dense(hidden_size, activation='relu'),
            tf.keras.layers.Dense(10)
        ])


# MLPug needs a TrainModel that outputs the loss
class TrainModel(tf.keras.Model):
    def __init__(self, classifier, global_batch_size):
        super(TrainModel, self).__init__()

        self.classifier = classifier
        self.global_batch_size = global_batch_size

        self.loss_func = tf.keras.losses.SparseCategoricalCrossentropy(
            from_logits=True,
            reduction=tf.keras.losses.Reduction.SUM)

    def call(self, batch_data, evaluate_settings, inference_mode=None):
        images, true_labels = batch_data

        logits = self.classifier(images, training=not inference_mode)
        loss = self.loss_func(true_labels, logits)/self.global_batch_size

        return loss


def train_model(args, logger):

    distributed = args.distributed

    # ########### EXPERIMENT SETUP ############
    tf.random.set_seed(args.seed)  # For reproducibility

    num_gpus_available = len(tf.config.list_physical_devices('GPU'))

    if distributed:
        if num_gpus_available < 1:
            logger.error(f"No GPUs available for data distributed training over multiple GPUs")
            return

        num_devices = args.num_devices if args.num_devices > 0 else num_gpus_available
        if num_devices > num_gpus_available:
            logger.warn(f"Number of requested GPUs is lower than available GPUs, "
                        f"limiting training to {num_gpus_available} GPUS")
            num_devices = num_gpus_available

        devices = [f"/gpu:{i}" for i in range(num_devices)]
        strategy = tf.distribute.MirroredStrategy(devices=devices)

        global_batch_size = args.batch_size * strategy.num_replicas_in_sync
    else:
        device = "/device:GPU:0" if num_gpus_available > 0 else "/CPU:0"

        strategy = tf.distribute.OneDeviceStrategy(device=device)
        global_batch_size = args.batch_size

    # ########################################

    # ########## SETUP BATCH DATASETS ##########
    logger.info(f"Global batch size: {global_batch_size}")

    train_images, train_labels, test_images, test_labels = load_data()

    training_dataset = tf.data.Dataset.from_tensor_slices((train_images, train_labels)).batch(global_batch_size)

    # Using the test set as a validation set, just for demonstration purposes
    validation_dataset = tf.data.Dataset.from_tensor_slices((test_images, test_labels)).batch(global_batch_size)

    num_batches_per_epoch = int(training_dataset.cardinality())
    if distributed:
        # Distribute training and validation set
        training_dataset = strategy.experimental_distribute_dataset(training_dataset)
        validation_dataset = strategy.experimental_distribute_dataset(validation_dataset)

    # ##########################################

    # Build within the strategy scope
    with strategy.scope():
        # ############ BUILD THE MODEL #############

        classifier = build_model(args.hidden_size)

        train_model = TrainModel(classifier, global_batch_size)
        # ##########################################

        # ############ SETUP OPTIMIZER #############
        optimizer = tf.keras.optimizers.Adam()
        # ##########################################

    # ############# SETUP TRAINING ##############
    print(f"Element spec : \n{training_dataset.element_spec}")
    trainer = mlp.trainers.DefaultTrainer(optimizers=optimizer,
                                          model_components=classifier,
                                          distribution_strategy=strategy,
                                          # See issue https://github.com/tensorflow/tensorflow/issues/29911
                                          batch_data_signature=training_dataset.element_spec)

    model_hyper_parameters = {
        "hidden_size": args.hidden_size
    }

    callbacks = create_callbacks_for(trainer,
                                     args.experiment_name,
                                     model_hyper_parameters,
                                     strategy,
                                     validation_dataset,
                                     args.progress_log_period)

    manager = mlp.trainers.TrainingManager(trainer,
                                           training_dataset,
                                           num_epochs=args.num_epochs,
                                           num_batches_per_epoch=num_batches_per_epoch,
                                           callbacks=callbacks)

    trainer.set_training_model(train_model)
    # ##########################################

    # ################# START! #################
    manager.start_training()
    # ##########################################


def test_model(model_checkpoint_filename, logger, device=None):
    _, _, test_images, test_labels = load_data()

    if device is None:
        device = "/CPU:0"

    logger.info(f'Loading model checkpoint ...')
    with open(model_checkpoint_filename, 'rb') as f:
        checkpoint = pickle.load(f)

    with tf.device(device):
        # hyper_parameters contains 'hidden_size'
        classifier = build_model(**checkpoint['hyper_parameters'])
        with h5py.File(checkpoint['model'], 'r') as f:
            hdf5_format.load_weights_from_hdf5_group(f, classifier.layers)

        image = test_images[0]
        real_label = test_labels[0]

        logits = classifier(tf.expand_dims(image, 0))
        probabilities = tf.nn.softmax(logits)

        predicted_label = tf.math.argmax(probabilities, axis=-1)

        logger.info(f"real label = {real_label}, predicted label = {tf.squeeze(predicted_label)}\n")


if __name__ == '__main__':
    # ############# SETUP LOGGING #############
    mlp.logging.use_fancy_colors()

    logger_name = os.path.basename(__file__)
    logger = get_logger(logger_name)
    # ########################################

    # ############## PARSE ARGS ##############
    parser = base_argument_set()

    parser.parse_args()

    args = parser.parse_args()

    describe_args(args, logger)

    # ############## TRAIN MODEL ##############
    train_model(args, logger)

    # ######### USE THE TRAINED MODEL ##########
    sys.stdout.write("\n\n\n\n")
    sys.stdout.flush()

    logger.info("Using the classifier ...")

    num_gpus_available = len(tf.config.list_physical_devices('GPU'))
    device = "/device:GPU:0" if num_gpus_available > 0 else "/CPU:0"

    model_checkpoint_filename = f'../trained-models/{args.experiment_name}-best-model-checkpoint.m-ckp'
    test_model(model_checkpoint_filename, device=device, logger=logger)



import torch_xla.core.xla_model as xm

from mlpug.pytorch.trainers import \
    Trainer, \
    DefaultTrainer as DefaultTrainerPyTorch, \
    TrainingManager as TrainingManagerPyTorch

from mlpug.mlpug_exceptions import TrainerInvalidException


class TrainingManager(TrainingManagerPyTorch):

    def _training_ended(self):
        if self.is_distributed:
            # Wait for all processes to finish
            xm.rendezvous("mlpug-training-ended")


class DefaultTrainer(DefaultTrainerPyTorch):

    def __init__(self, *args, use_mixed_precision=False, name="DefaultTrainer", **kwargs):
        if use_mixed_precision:
            raise TrainerInvalidException("Mixed precision training not supported with XLA devices")

        super(DefaultTrainer, self).__init__(*args,
                                             use_mixed_precision=use_mixed_precision,
                                             name=name,
                                             **kwargs)

    def _back_propagate_from(self, loss, last_chunk=False):
        super()._back_propagate_from(loss, last_chunk=last_chunk)
        # Required when evaluating batch chunks for gradient accumulation
        xm.mark_step()

    def _update_model_parameters(self):
        for optimizer in self.get_optimizers().values():
            xm.optimizer_step(optimizer)



from mlpug.base import Base
from mlpug.mlpug_exceptions import StateInvalidException, InvalidParametersException


class SlidingWindow(Base):

    def __init__(self, length=None, init_window_values=None, name=None, state=None, **kwargs):
        """

        Can raise StateInvalidException or InvalidParametersException

        :param length:
        :type length:
        :param init_window_values:
        :type init_window_values:
        :param state:
        :type state:
        :param kwargs:
        :type kwargs:
        """
        if state:
            try:
                length = state["length"]
                init_window_values = state["window"]
                name = state["name"] if "name" in state else None
            except Exception as e:
                raise StateInvalidException() from e

        logger_name = self.__class__.__name__
        if name is not None:
            logger_name = f"{logger_name}[{name}]"

        super().__init__(pybase_logger_name=logger_name, **kwargs)

        self.name = name

        if not length:
            raise InvalidParametersException("No length value given, unable to instantiate SlidingWindow object.")

        self._log.debug(f"Window length : {length}")

        self._centre_not_exact_warning_given = False

        self.length = length
        self.center_index = self.length // 2

        if hasattr(init_window_values, "__getitem__"):
            self._log.debug("Using given initial values to fill the window as much as possible.")

            iwv_len = len(init_window_values)

            if iwv_len < self.length:
                self._log.debug(f"Given initial window values list is shorter than window: {iwv_len}")

                if iwv_len < 1:
                    self._log.warn("Unexpected : The given initial window values list is empty, continuing anyway ...")
            elif iwv_len > self.length:
                iwv_len = self.length
                self._log.debug(f"Given initial window values list is longer than window, "
                                f"only using the last {iwv_len} values to fill the whole window")
            else:
                self._log.debug(f"Given initial window values list matches the window length perfectly")

            self.window = init_window_values[-iwv_len:][:]
        else:
            self.window = list()

    def get_state(self):
        return {
            "window": self.window,
            "length": self.length,
            "name": self.name
        }

    def flush(self):
        self.window.clear()

    def center(self):

        if not self.is_filled():
            self._log.error("Window not completely filled yet, unable to get center")
            return None

        if self.length % 2 != 1 and not self._centre_not_exact_warning_given:
            self._log.warn("Given window length is even, centre is not exact")
            self._centre_not_exact_warning_given = True

        return self.window[self.center_index]

    def last(self):
        return self.window[-1]

    def first(self):
        return self.window[0]

    def is_filled(self):
        return len(self.window) >= self.length

    def is_empty(self):
        return len(self.window) == 0

    def slide(self, value):
        if self.is_filled():
            self.window = self.window[1:]

        self.window.append(value)



import abc

from mlpug.base import Base

from mlpug.utils.utils import get_value_at


class Callback(Base, metaclass=abc.ABCMeta):

    def __init__(self, name, base_logs_path="current", **kwargs):
        super(Callback, self).__init__(pybase_logger_name=name, **kwargs)

        self.name = name

        self.base_log_path = base_logs_path

        self.training_manager = None
        self.trainer = None
        self.model_components = None
        self.optimizers = None

    def set_training_manager(self, manager):
        self.training_manager = manager
        self.trainer = manager.get_trainer()
        self.model_components = self.trainer.get_model_components()
        self.optimizers = self.trainer.get_optimizers()

        return True

    def get_name(self):
        return self.name

    def get_state(self):
        """

        :return: state, success (True or False)
        """
        return None, True

    def set_state(self, state):
        """

        :param state:
        :return: success (True or False)
        """
        return True

    def on_training_start(self,
                          num_epochs,
                          num_batches_per_epoch,
                          start_epoch,
                          start_batch,
                          start_update_iter):
        """

        :param num_epochs:
        :param num_batches_per_epoch:
        :param start_epoch:
        :param start_batch:
        :param start_update_iter:

        :return: success (True or False)
        """
        return True

    def on_epoch_start(self, logs):
        """

        :param logs:

        :return: success (True or False)
        """
        return True

    def on_batch_training_start(self, training_batch, logs):
        """

        :param training_batch:
        :param logs:

        :return: success (True or False)
        """
        return True

    def on_batch_training_failed(self, exception, logs):
        """

        :param exception:
        :param logs:

        :return: success (True or False)
        """
        return True

    def on_batch_training_completed(self, training_batch, logs):
        """

        :param training_batch:
        :param logs:

        :return: success (True or False)
        """
        return True

    def on_epoch_completed(self, logs):
        """

        :param logs:

        :return: success (True or False)
        """
        return True

    def on_training_ended(self, stopped_early, stopped_on_error, interrupted, callback_calls_success):
        """

        :param stopped_early:
        :param stopped_on_error:
        :param interrupted:
        :param callback_calls_success:

        :return: success (True or False)
        """
        return True

    def _get_logs_base(self, logs):
        if self.base_log_path is None:
            return logs
        else:
            return get_value_at(self.base_log_path, logs)




# This is from the PyTorch Chatbot tutorial https://pytorch.org/tutorials/beginner/chatbot_tutorial.html
# but is not PyTorch specific

class Voc:
    def __init__(self, name, PAD_token=0, SOS_token=1, EOS_token=2):
        self.name = name
        self.PAD_token = PAD_token
        self.SOS_token = SOS_token
        self.EOS_token = EOS_token

        self.trimmed = False
        self.word2index = {}
        self.word2count = {}
        self.index2word = {self.PAD_token: "PAD", self.SOS_token: "SOS", self.EOS_token: "EOS"}
        self.num_words = 3  # Count SOS, EOS, PAD

    def addSentence(self, sentence):
        for word in sentence.split(' '):
            self.addWord(word)

    def addWord(self, word):
        if word not in self.word2index:
            self.word2index[word] = self.num_words
            self.word2count[word] = 1
            self.index2word[self.num_words] = word
            self.num_words += 1
        else:
            self.word2count[word] += 1

    # Remove words below a certain count threshold
    def trim(self, min_count):
        if self.trimmed:
            return
        self.trimmed = True

        keep_words = []

        for k, v in self.word2count.items():
            if v >= min_count:
                keep_words.append(k)

        print('keep_words {} / {} = {:.4f}'.format(
            len(keep_words), len(self.word2index), len(keep_words) / len(self.word2index)
        ))

        # Reinitialize dictionaries
        self.word2index = {}
        self.word2count = {}
        self.index2word = {self.PAD_token: "PAD", self.SOS_token: "SOS", self.EOS_token: "EOS"}
        self.num_words = 3 # Count default tokens

        for word in keep_words:
            self.addWord(word)



import time

import os
import sys

import tensorflow as tf
import tensorflow_datasets as tfds

from mlpug.examples.chatbot.conversation_dataset import load_sentence_pair_data

from mlpug.examples.chatbot.tensorflow.original_transformer_tutorial.model_data_generation import \
    create_length_filter_func, \
    create_chatbot_tf_encode_func

from mlpug.examples.chatbot.tensorflow.original_transformer_tutorial.training import \
    CustomSchedule, \
    create_masks, \
    loss_function

from mlpug.examples.chatbot.tensorflow.original_transformer_tutorial.transformer import Transformer

# ############# SETUP ###############
REMOTE_DEBUG = False

experiment_name = "large-transformer-01122020-4-layers"

data_set_path = "./mlpug/examples/chatbot/data/"

training_set_file = "training-1606433382-cmdc-sentence-pairs-with-voc-max-len-40-min-word-occurance-3-26112020.pickle"
validation_set_file = "validation-1606433382-cmdc-sentence-pairs-with-voc-max-len-40-min-word-occurance-3-26112020.pickle"

num_layers = 4
d_model = 768
dff = 3072
num_heads = 12

dropout_rate = 0.1

log_interval = 20

MAX_LENGTH = 40

BUFFER_SIZE = 20000
BATCH_SIZE = 64

EPOCHS = 200
#####################################

if REMOTE_DEBUG:
    import pydevd
    pydevd.settrace('192.168.178.85', port=57491, stdoutToServer=True, stderrToServer=True)


# ########### SETUP DATA ############

def create_dataset_generator(pairs):

    def generator():
        for pair in pairs:
            yield tuple(pair)

    return generator

train_examples, _unused_ = load_sentence_pair_data(os.path.join(data_set_path, training_set_file))
val_examples, _unused_ = load_sentence_pair_data(os.path.join(data_set_path, validation_set_file))

all_training_sentences = []
for pair in train_examples:
    all_training_sentences += pair

tokenizer = tfds.deprecated.text.SubwordTextEncoder.build_from_corpus(
    all_training_sentences, target_vocab_size=2**13)

vocab_size = tokenizer.vocab_size + 2

tf_encode = create_chatbot_tf_encode_func(tokenizer)
filter_max_length = create_length_filter_func(MAX_LENGTH)

train_dataset = tf.data.Dataset.from_generator(
    create_dataset_generator(train_examples),
    (tf.string, tf.string),
    (tf.TensorShape([]), tf.TensorShape([])))

train_dataset = train_dataset.map(tf_encode)
train_dataset = train_dataset.filter(filter_max_length)
# cache the dataset to memory to get a speedup while reading from it.
train_dataset = train_dataset.cache()
train_dataset = train_dataset.shuffle(BUFFER_SIZE).padded_batch(BATCH_SIZE)
train_dataset = train_dataset.prefetch(tf.data.experimental.AUTOTUNE)


val_dataset = tf.data.Dataset.from_generator(
    create_dataset_generator(val_examples),
    (tf.string, tf.string),
    (tf.TensorShape([]), tf.TensorShape([])))

val_dataset = val_dataset.map(tf_encode)
val_dataset = val_dataset.filter(filter_max_length).padded_batch(BATCH_SIZE)

#####################################


# ######## SETUP OPTIMIZER ##########
learning_rate = CustomSchedule(d_model)

#learning_rate = 5e-3
optimizer = tf.keras.optimizers.Adam(learning_rate, beta_1=0.9, beta_2=0.98, epsilon=1e-9)

train_loss = tf.keras.metrics.Mean(name='train_loss')
train_accuracy = tf.keras.metrics.SparseCategoricalAccuracy(name='train_accuracy')
#####################################

# ########## SETUP MODEL ############
transformer = Transformer(num_layers, d_model, num_heads, dff,
                          vocab_size, vocab_size,
                          pe_input=vocab_size,
                          pe_target=vocab_size,
                          rate=dropout_rate)

checkpoint_path = os.path.join("../trained-models/", experiment_name)
#####################################


# The @tf.function trace-compiles train_step into a TF graph for faster
# execution. The function specializes to the precise shape of the argument
# tensors. To avoid re-tracing due to the variable sequence lengths or variable
# batch sizes (the last batch is smaller), use input_signature to specify
# more generic shapes.

train_step_signature = [
    tf.TensorSpec(shape=(None, None), dtype=tf.int64),
    tf.TensorSpec(shape=(None, None), dtype=tf.int64),
]


@tf.function(input_signature=train_step_signature)
def train_step(inp, tar):
    tar_inp = tar[:, :-1]
    tar_real = tar[:, 1:]

    enc_padding_mask, combined_mask, dec_padding_mask = create_masks(inp, tar_inp)

    with tf.GradientTape() as tape:
        predictions, _ = transformer(inp, tar_inp,
                                     True,
                                     enc_padding_mask,
                                     combined_mask,
                                     dec_padding_mask)
        loss = loss_function(tar_real, predictions)

    gradients = tape.gradient(loss, transformer.trainable_variables)
    optimizer.apply_gradients(zip(gradients, transformer.trainable_variables))

    train_loss(loss)
    train_accuracy(tar_real, predictions)


ckpt = tf.train.Checkpoint(transformer=transformer,
                           optimizer=optimizer)

ckpt_manager = tf.train.CheckpointManager(ckpt, checkpoint_path, max_to_keep=1)

# if a checkpoint exists, restore the latest checkpoint.
if ckpt_manager.latest_checkpoint:
    ckpt.restore(ckpt_manager.latest_checkpoint)
    print('Latest checkpoint restored!!')

for epoch in range(EPOCHS):
    start = time.time()

    train_loss.reset_states()
    train_accuracy.reset_states()

    for (batch, (inp, tar)) in enumerate(train_dataset):
        train_step(inp, tar)

        if batch % log_interval == 0:
            print('Epoch {} Batch {} Loss {:.4f} Accuracy {:.4f}'.format(
                epoch + 1, batch, train_loss.result(), train_accuracy.result()))
            sys.stdout.flush()

    if (epoch + 1) % 5 == 0:
        ckpt_save_path = ckpt_manager.save()
        print('Saving checkpoint for epoch {} at {}'.format(epoch + 1,
                                                            ckpt_save_path))

    print('Epoch {} Loss {:.4f} Accuracy {:.4f}'.format(epoch + 1,
                                                        train_loss.result(),
                                                        train_accuracy.result()))

    print('Time taken for 1 epoch: {} secs\n'.format(time.time() - start))



import sys

import torch
import torch.distributed as dist

from mlpug.pytorch.trainers.callbacks import Callback

from mlpug.mlpug_exceptions import CallbackInvalidException


class EmptyCudaCache(Callback):

    def __init__(self, on_batch_training_complete=True, on_epoch_complete=True, name="EmptyCudaCache", **kwargs):
        super(EmptyCudaCache, self).__init__(name=name, **kwargs)

        self._on_batch_training_complete = on_batch_training_complete
        self._on_epoch_complete = on_epoch_complete

        self._log.debug(f"Will empty CUDA cache on batch training complete : {self._on_batch_training_complete}")
        self._log.debug(f"Will empty CUDA cache on epoch complete : {self._on_epoch_complete}")

    def on_batch_training_completed(self, training_batch, logs):
        if not self._on_batch_training_complete:
            return True

        torch.cuda.empty_cache()

        return True

    def on_epoch_completed(self, logs):
        if not self._on_epoch_complete:
            return True

        torch.cuda.empty_cache()

        return True


class LogCudaMemory(Callback):

    def __init__(self,
                 devices=None,
                 on_batch_training_start=False,
                 on_batch_training_complete=True,
                 name="LogCudaMemory",
                 **kwargs):
        """

        TODO: add memory statistics to log

        :param devices: list of indices of devices to log memory for
        :param on_batch_training_start:
        :param on_batch_training_complete:
        :param name:
        """

        super(LogCudaMemory, self).__init__(name=name, **kwargs)

        try:
            self._pynvml = __import__('pynvml')
        except Exception as e:
            raise CallbackInvalidException("The pynvml is not installed, please do "
                                           "`pip install pynvml` in your python environment")

        if devices is None:
            num_devices = dist.get_world_size() if dist.is_initialized() else 1
            devices = range(num_devices)
        else:
            num_devices = len(devices)

        self._log.debug(f"Devices to log memory usage for : {list(devices)}")

        self._num_devices = num_devices
        self._devices = devices

        self._on_batch_training_start = on_batch_training_start
        self._on_batch_training_complete = on_batch_training_complete

        self._max_memory_used = []

        self._pynvml.nvmlInit()

    def on_batch_training_start(self, training_batch, logs):
        if not self._on_batch_training_start:
            return True

        self._log_memory()

        return True

    def on_batch_training_completed(self, training_batch, logs):
        if not self._on_batch_training_complete:
            return True

        self._log_memory()

        return True

    def _log_memory(self):
        sys.stdout.write("\n")
        sys.stdout.write("|----------|----------------|----------------|----------------| -------------- |\n")
        sys.stdout.write("| Device # | Free           | Used           | Max. Used      | Total          |\n")
        sys.stdout.write("|----------|----------------|----------------|----------------| -------------- |\n")
        sys.stdout.flush()
        for device_index in self._devices:
            device_handle = self._pynvml.nvmlDeviceGetHandleByIndex(device_index)
            mem_info = self._pynvml.nvmlDeviceGetMemoryInfo(device_handle)

            total_memory = self._to_mb(mem_info.total)
            memory_used = self._to_mb(mem_info.used)
            memory_free = self._to_mb(mem_info.free)

            if len(self._max_memory_used) < self._num_devices:
                self._max_memory_used += [memory_used]
            else:
                self._max_memory_used[device_index] = max(memory_used, self._max_memory_used[device_index])

            max_memory_used = self._max_memory_used[device_index]

            memory_values_str_format = "| {:<8} | {:<12}MB | {:<12}MB | {:<12}MB | {:<12}MB |\n"
            memory_values_str = memory_values_str_format.format(device_index,
                                                                memory_free,
                                                                memory_used,
                                                                max_memory_used,
                                                                total_memory)

            sys.stdout.write(memory_values_str)

        sys.stdout.write("|----------|----------------|----------------|----------------| -------------- |\n")
        sys.stdout.write("\n")
        sys.stdout.flush()

    def _to_mb(self, num_bytes):
        return int(round(num_bytes/(1024*1024), 0))



from enum import Enum

import math

from mlpug.trainers.callbacks.callback import Callback
from mlpug.mlpug_exceptions import CallbackInvalidException
from mlpug.utils import has_key, get_key_paths, SlidingWindow

from mlpug.evaluation import MetricEvaluatorBase

import basics.base_utils as _

from mlpug.utils import get_value_at, set_value_at


class MetricsLoggingMode(Enum):
    # TODO : Add BATCH_METRICS_ONLY Option
    BATCH_AND_WINDOW_AVERAGE_METRICS = "BATCH_AND_WINDOW_AVERAGE_METRICS"
    WINDOW_AVERAGE_METRICS = "WINDOW_AVERAGE_METRICS"
    WHOLE_DATASET_METRICS = "WHOLE_DATASET_METRICS"

    @staticmethod
    def will_log_window_average_metrics(mode):
        return mode in {MetricsLoggingMode.BATCH_AND_WINDOW_AVERAGE_METRICS,
                        MetricsLoggingMode.WINDOW_AVERAGE_METRICS}

    def __str__(self):
        return str(self.value)


class MetricsLoggerBase(Callback):

    def __init__(self,
                 metric_evaluator,
                 dataset_name,
                 batch_level,
                 logging_mode,
                 dataset=None,
                 evaluate_settings=None,
                 batch_averaging_window=None,
                 log_condition_func=None,
                 sliding_window_factory=None,
                 name=None,
                 **kwargs):
        """

        Different modes of operation:

        Batch level: After every batch training iteration:
            Calculate metrics over a sliding window
                log metrics per batch and over window   batch_level = True, mode = BATCH_AND_WINDOW_AVERAGE_METRICS
                log metrics over window only            batch_level = True, mode = WINDOW_AVERAGE_METRICS
            Calculate metrics over whole data set       batch_level = True, mode = WHOLE_DATASET_METRICS

        Epoch level: After every training epoch:
            Calculate metrics over whole data set       batch_level = False, mode = WHOLE_DATASET_METRICS


        `log_condition_func` can be used to control if metrics are calculated

        :param dataset_name:
        :type dataset_name:

        :param batch_level: See general description above
        :type batch_level:

        :param logging_mode: See general description above
        :type logging_mode:

        :param dataset: Dataset, where the items are batches, to calculate the metrics with
        :type dataset:

        :param metric_evaluator: MetricEvaluator instance, or other object with the same public interface
        :type metric_evaluator:

        :param evaluate_settings: Default evaluation settings to use, when no evaluate_settings are available in the
                                  logs object provided to the callback methods
        :type evaluate_settings:

        :param batch_averaging_window: Window length in batches, over which the average metric must be calculated
        :type batch_averaging_window: Int

        :param sliding_window_factory: Optional. Function to create SlidingWindow instances

        :param name:
        :type name:
        """
        # print(f"mlpug.MetricsLoggerBase(kwargs={kwargs})")

        if name is None:
            name = self.__class__.__name__

            logging_level = "Batch level" if batch_level else "Epoch level"
            name = f"{name}[{dataset_name} dataset][{logging_level}][{str(logging_mode)}]"

        super().__init__(name=name, **kwargs)

        self._dataset = dataset
        self._dataset_name = dataset_name

        self._metric_evaluator = metric_evaluator

        self._batch_level = batch_level

        self._logging_mode = logging_mode

        self._evaluate_settings = evaluate_settings

        self._batch_averaging_window = batch_averaging_window

        self._log_condition_func = log_condition_func or (lambda logs, dataset_batch: True)

        self._sliding_window_factory = SlidingWindow if not callable(sliding_window_factory) else sliding_window_factory

        self._name = name

        self._metric_windows = {}

        self._validate()

        if not self.instance_valid():
            raise CallbackInvalidException(self._name)

    def get_state(self):
        """

        :return: state, success (True or False)
        """
        return {
                    "metric_windows": {metric_path: window.get_state()
                                       for metric_path, window in self._metric_windows.items()}
               }, True

    def set_state(self, state):
        """

        :param state:
        :return: success (True or False)
        """

        if not self._check_state(state):
            self._log.error('Given state is invalid, unable to set state')
            return False

        self._metric_windows = {metric_path: self._sliding_window_factory(state=window_state)
                                for metric_path, window_state in state["metric_windows"].items()}

        return True

    def on_training_start(self,
                          num_epochs,
                          num_batches_per_epoch,
                          start_epoch,
                          start_batch,
                          start_update_iter):

        will_log_window_average_metrics = MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode)

        if not self.instance_valid():
            if will_log_window_average_metrics:
                self._log.error(f"{self} is not valid, unable to set up averaging window ... ")
            else:
                self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if will_log_window_average_metrics and \
                (self._batch_averaging_window is None or self._batch_averaging_window <= 0):
            self._log.error(f"A valid batch processing window is required for "
                            f"metric logging mode ({self._logging_mode}, the {self} will not function")
            self._valid = False

            return False

        if will_log_window_average_metrics:
            self._init_metric_windows()

        return True

    def on_batch_training_completed(self, dataset_batch, logs):
        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if not self._batch_level:
            return True

        if not self._log_condition_func(logs=logs, dataset_batch=dataset_batch):
            return True

        self._init_logs(logs)

        if self._logging_mode is MetricsLoggingMode.WHOLE_DATASET_METRICS:
            return self._calc_whole_dataset_metrics(logs, f"{self._dataset_name}.dataset")
        else:
            current = self._get_logs_base(logs)

            batch_metrics = {}
            if not self._calc_batch_metric_data_from(dataset_batch, batch_metrics, logs):
                return False

            base_path = f"{self._dataset_name}.batch"
            dataset_batch_logs = get_value_at(base_path, current)

            # Merge in new batch level results
            dataset_batch_logs = {**dataset_batch_logs, **batch_metrics}
            if self._logging_mode is MetricsLoggingMode.BATCH_AND_WINDOW_AVERAGE_METRICS:
                set_value_at(base_path, current, dataset_batch_logs)

            metric_names = self._metric_evaluator.get_metric_names()
            metric_paths = get_key_paths(dataset_batch_logs,
                                         keys_to_consider=metric_names,
                                         keys_not_to_consider=["auxiliary_results"])

            self._update_metrics_windows_for(metric_paths, dataset_batch_logs, base_path=base_path)

            # gather all window data
            batch_metrics_lists = {p: s.window for p, s in self._metric_windows.items()}
            if not self._reduce(batch_metrics_lists, current[self._dataset_name]['window_average']):
                return False

            return True

    def on_epoch_completed(self, logs):
        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if self._batch_level:
            return True

        if not self._log_condition_func(logs=logs, dataset_batch=None):
            return True

        self._init_logs(logs)

        return self._calc_whole_dataset_metrics(logs, f"{self._dataset_name}.dataset")

    def _init_logs(self, logs):
        current = self._get_logs_base(logs)
        if self._dataset_name not in current:
            current[self._dataset_name] = {}

        dataset_metrics = current[self._dataset_name]
        for metric_level in ['batch', 'window_average', 'dataset']:
            if metric_level not in dataset_metrics:
                dataset_metrics[metric_level] = {}

    def _get_current_evaluate_settings(self, logs):
        # Use custom settings if available, else use default settings

        evaluate_settings = get_value_at('evaluate_settings', logs, warn_on_failure=False)
        if evaluate_settings is None:
            evaluate_settings = self._evaluate_settings

        return evaluate_settings

    def _calc_whole_dataset_metrics(self, logs, log_path):

        current = self._get_logs_base(logs)
        metrics_log = get_value_at(log_path, current)

        evaluate_settings = self._get_current_evaluate_settings(logs)

        return self._metric_evaluator.calc_dataset_metrics_for(self._dataset,
                                                               metrics_log,
                                                               evaluate_settings=evaluate_settings,
                                                               dataset_name=self._dataset_name)

    def _init_metric_windows(self, reset=False):
        if reset:
            self._metric_windows = {}

        # Do nothing if metric windows are already available
        self._metric_windows = self._metric_windows or {}

        return True

    def _calc_batch_metric_data_from(self, batch, batch_metrics, logs):
        evaluate_settings = self._get_current_evaluate_settings(logs)

        model_output = None
        if self._dataset is None:
            current = self._get_logs_base(logs)
            loss = get_value_at(f"{self._dataset_name}.batch.loss", current)
            auxiliary_results = get_value_at(f"{self._dataset_name}.batch.auxiliary_results",
                                             current,
                                             warn_on_failure=False)

            model_output = {
                'loss': loss,
                'auxiliary_results': auxiliary_results
            }

        return self._metric_evaluator.calc_batch_metrics_for(batch,
                                                             batch_metrics,
                                                             evaluate_settings=evaluate_settings,
                                                             model_output=model_output)

    def _update_metrics_windows_for(self, metric_paths, batch_metrics, base_path):
        for metric_path in metric_paths:
            metric_value = get_value_at(metric_path, batch_metrics)

            full_metric_path = f"{base_path}.{metric_path}"
            sliding_window = self._metric_windows[metric_path] if metric_path in self._metric_windows else None
            if sliding_window is None:
                self._log.debug(f"Creating sliding window for {full_metric_path}")

                sliding_window = self._sliding_window_factory(length=self._batch_averaging_window, name=full_metric_path)
                self._metric_windows[metric_path] = sliding_window

            sliding_window.slide(metric_value)

    def _reduce(self, batch_metric_data_lists, reduced_metrics_log):

        return self._metric_evaluator.reduce(batch_metric_data_lists,
                                             reduced_metrics_log,
                                             dataset_name=self._dataset_name)

    def _check_state(self, state):
        state_attributes = ['metric_windows']

        for attr in state_attributes:
            if not has_key(state, attr):
                self._log.error(f"Given state does not have a value for {attr}, state is invalid")
                return False

        return True

    def _validate(self):
        self._valid = True

        if not hasattr(self._dataset, '__iter__'):
            if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode):
                self._log.debug("Training batch evaluation results will be used to calculate and log metrics.")
            else:
                self._log.error(f"No valid dataset provided to calculated metrics on, "
                                f"the {self} will not function")
                self._valid = False

        if not MetricEvaluatorBase.is_valid(self._metric_evaluator):
            self._log.error(f"The given metric evaluator is not valid, the {self} will not function : "
                            f"{self._metric_evaluator}")
            self._valid = False

        if not type(self._batch_level) is bool:
            self._log.error(f"Batch level must be True or False, given value is: {self._batch_level}\n"
                            f"the {self} will not function")
            self._valid = False

        if self._logging_mode not in MetricsLoggingMode:
            self._log.error(f"Metric logging mode unknown ({self._logging_mode})\n"
                            f"the {self} will not function")
            self._valid = False

        if not self._batch_level and self._logging_mode is not MetricsLoggingMode.WHOLE_DATASET_METRICS:
            self._log.error(f"For epoch level metric logging, the metric logging can only be "
                            f"MetricsLoggingMode.WHOLE_DATASET_METRICS. value is :{self._logging_mode}\n"
                            f"the {self} will not function")
            self._valid = False

        if self._log_condition_func is not None and not callable(self._log_condition_func):
            self._log.error(f"The log condition function must be callable, "
                            f"metric logging mode ({self._logging_mode}, the {self} will not function")
            self._valid = False


class TrainingMetricsLogger(MetricsLoggerBase):
    """
    By default, gets already calculated loss and auxiliary results from logs

    TODO : needs testing
    TODO : what to do for batch_level=false? Allow usage of window average of metrics to be used for epoch level?
    """

    def __init__(self,
                 metric_evaluator,
                 batch_level=True,
                 logging_mode=MetricsLoggingMode.BATCH_AND_WINDOW_AVERAGE_METRICS,
                 **kwargs):

        # print(f"mlpug.TrainingMetricsLogger(kwargs={kwargs})")

        super().__init__(
            metric_evaluator=metric_evaluator,
            dataset_name='training',
            batch_level=batch_level,
            logging_mode=logging_mode,
            **kwargs)

    def on_training_start(self,
                          num_epochs,
                          num_batches_per_epoch,
                          start_epoch,
                          start_batch,
                          start_update_iter):

        if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode):
            if self._batch_averaging_window is None:
                self._log.debug(f"Set batch_averaging_window to number of batches per epoch : {num_batches_per_epoch}")
                self._batch_averaging_window = num_batches_per_epoch

            if self._batch_averaging_window == math.inf:
                self._log.error("The batch average window is infinite, unable to calculate window average. "
                                "Please set a finite batch_averaging_window during construction of this "
                                "TrainingMetricsLogger.")
                self._valid = False
                return False

        return super().on_training_start(num_epochs, num_batches_per_epoch, start_epoch, start_batch, start_update_iter)


class TestMetricsLogger(MetricsLoggerBase):
    """
    Child class must implement _evaluate_loss method

    TODO : We should implement a specialization for the most common case of logging the (window/dataset) loss average
    """

    def __init__(self,
                 dataset,
                 dataset_name,
                 metric_evaluator,
                 batch_level=True,
                 logging_mode=None,
                 **kwargs):

        if logging_mode is None:
            logging_mode = MetricsLoggingMode.BATCH_AND_WINDOW_AVERAGE_METRICS \
                if batch_level else MetricsLoggingMode.WHOLE_DATASET_METRICS

        super().__init__(dataset=dataset,
                         dataset_name=dataset_name,
                         metric_evaluator=metric_evaluator,
                         batch_level=batch_level,
                         logging_mode=logging_mode,
                         **kwargs)

        self._dataset_iterator = None

    def on_training_start(self,
                          num_epochs,
                          num_batches_per_epoch,
                          start_epoch,
                          start_batch,
                          start_update_iter):

        if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode):
            if self._batch_averaging_window is None:
                try:
                    self._batch_averaging_window = len(self._dataset)
                    if self._batch_averaging_window == math.inf:
                        self._log.error(f"The batch average window of the {self._dataset_name} data set is infinite, "
                                        f"unable to calculate window average. "
                                        f"Please set a finite batch_processing_window during construction of "
                                        f"this TestMetricsLogger.")
                        self._valid = False
                    else:
                        self._log.debug(f"Set batch averaging window to number of batches in "
                                        f"{self._dataset_name} data set : {self._batch_averaging_window}")
                except Exception as e:
                    _.log_exception(self._log, f"Unable to assess data set length to set the batch averaging window, "
                                               f"{self} will not function", e)
                    self._valid = False

        return super().on_training_start(num_epochs, num_batches_per_epoch, start_epoch, start_batch, start_update_iter)

    def on_epoch_start(self, logs):
        """

        :param logs:

        :return: success (True or False)
        """
        success = super().on_epoch_start(logs)

        if not success:
            self._log.error('A problem occurred, will not continue executing this hook')
            return success

        if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode) and \
                self._dataset_iterator is None:
            self._dataset_iterator = iter(self._dataset)

        return True

    def on_batch_training_completed(self, training_batch, logs):
        current = self._get_logs_base(logs)
        training_iter = current["global_iter"]

        if self._logging_mode is MetricsLoggingMode.WHOLE_DATASET_METRICS:
            return super().on_batch_training_completed(training_batch, logs)
        else:
            try:
                dataset_batch = next(self._dataset_iterator)
            except StopIteration:
                self._log.debug(f"[Iteration {training_iter}] Rewinding the {self._dataset_name} data set")
                self._dataset_iterator = iter(self._dataset)
                dataset_batch = next(self._dataset_iterator)

            return super().on_batch_training_completed(dataset_batch, logs)

    def _validate(self):
        if not hasattr(self._dataset, '__iter__'):
            self._log.error(f"The given dataset {str(self._dataset)} is not iterable, the {self} will not function")
            self._valid = False

        super()._validate()



import os

import tensorflow as tf

from mlpug.evaluation import default_metric_reducer_func, MetricEvaluatorBase

from mlpug.utils import is_chunkable

from mlpug.base import Base
from basics.logging import get_logger

logger = get_logger(os.path.basename(__file__))


# ####### DEFAULT GATHER LOSS METHODS ########
class GatherLossSimple(Base):

    def __init__(self, distribution_strategy=None, requester=None):
        name = "GatherLossSimple"
        if requester is not None:
            name += f'[{requester}]'

        super(GatherLossSimple, self).__init__(pybase_logger_name=name)

        self._distribution_strategy = distribution_strategy
        self.requester = requester

        self._gather_loss_func = None
        if self._distribution_strategy is not None:
            self._log.info(f"Using simple distributed gather loss function")
            self._gather_loss_func = self._gather_loss_distributed
        else:
            self._log.info(f"Using simple gather loss function")
            self._gather_loss_func = self._gather_loss

    def __call__(self, *args, **kwargs):
        return self._gather_loss_func(*args, **kwargs)

    def _gather_loss(self, loss, **kwargs):
        loss = loss.numpy()
        return loss, loss, 1

    def _gather_loss_distributed(self, loss, **kwargs):
        loss = self._distribution_strategy.reduce(tf.distribute.ReduceOp.SUM, loss, axis=None)
        loss = loss.numpy()

        return loss, loss, 1


class GatherMaskedLoss(Base):

    def __init__(self, distribution_strategy=None, requester=None):
        name = "GatherMaskedLoss"
        if requester is not None:
            name += f'[{requester}]'

        super(GatherMaskedLoss, self).__init__(pybase_logger_name=name)

        self._distribution_strategy = distribution_strategy
        self.requester = requester

        self._gather_loss_func = None
        if self._distribution_strategy is not None:
            self._log.info(f"Using distributed gather masked loss function")
            self._gather_loss_func = self._gather_loss_distributed
        else:
            self._log.info(f"Using gather masked loss function")
            self._gather_loss_func = self._gather_loss

    def __call__(self, *args, **kwargs):
        return self._gather_loss_func(*args, **kwargs)

    def _gather_loss(self, auxiliary_results, **kwargs):
        loss_sum = auxiliary_results[0].item()
        num_samples = auxiliary_results[1].item()

        loss = loss_sum/num_samples

        return loss, loss_sum, num_samples

    def _gather_loss_distributed(self, auxiliary_results, **kwargs):
        loss_sum_per_replica = auxiliary_results[0]
        num_samples_per_replica = auxiliary_results[1]

        loss_sum = self._distribution_strategy.reduce(tf.distribute.ReduceOp.SUM, loss_sum_per_replica, axis=None)
        num_samples = self._distribution_strategy.reduce(tf.distribute.ReduceOp.SUM, num_samples_per_replica, axis=None)

        loss = loss_sum/num_samples

        return loss.numpy(), loss_sum.numpy(), num_samples.numpy()

# ############################################


class MetricEvaluator(MetricEvaluatorBase):

    def __init__(self, *args, distribution_strategy=None, batch_metric_funcs=None, name="MetricEvaluator", **kwargs):

        self.distribution_strategy = distribution_strategy

        if batch_metric_funcs is None:
            batch_metric_funcs = {
                "loss": GatherLossSimple(distribution_strategy, requester=name)
            }

        super().__init__(batch_metric_funcs, *args, name=name, **kwargs)

    def _create_default_model_evaluate_func(self):

        def evaluate_loss(batch, evaluate_settings=None):
            if is_chunkable(batch):
                # Get raw batch
                batch = batch[:]

            results = self._trainer.evaluate_loss(
                batch,
                inference_mode=True,
                evaluate_settings=evaluate_settings)

            return results

        return evaluate_loss



# Note : Optimized for chatbot usecase

import tensorflow as tf

import matplotlib.pyplot as plt

from mlpug.examples.chatbot.tensorflow.original_transformer_tutorial.training import create_masks


class Evaluator:
    # Note : only deterministic output

    def __init__(self, tokenizer, transformer, max_length):

        self.tokenizer = tokenizer
        self.transformer = transformer
        self.max_length = max_length

        self.start_token = [self.tokenizer.vocab_size]
        self.end_token = [self.tokenizer.vocab_size + 1]

    def __call__(self, inp_sentence):

        inp_sentence = self.start_token + self.tokenizer.encode(inp_sentence) + self.end_token
        encoder_input = tf.expand_dims(inp_sentence, 0)

        decoder_input = [self.tokenizer.vocab_size]
        output = tf.expand_dims(decoder_input, 0)

        for i in range(self.max_length):
            enc_padding_mask, combined_mask, dec_padding_mask = create_masks(
                encoder_input, output)

            # predictions.shape == (batch_size, seq_len, vocab_size)
            predictions, attention_weights = self.transformer(encoder_input,
                                                              output,
                                                              False,
                                                              enc_padding_mask,
                                                              combined_mask,
                                                              dec_padding_mask)

            # select the last word from the seq_len dimension
            predictions = predictions[:, -1:, :]  # (batch_size, 1, vocab_size)

            predicted_id = tf.cast(tf.argmax(predictions, axis=-1), tf.int32)

            # return the result if the predicted_id is equal to the end token
            if predicted_id == self.tokenizer.vocab_size + 1:
                return tf.squeeze(output, axis=0), attention_weights

            # concatenate the predicted_id to the output which is given to the decoder
            # as its input.
            output = tf.concat([output, predicted_id], axis=-1)

        return tf.squeeze(output, axis=0), attention_weights


class AttentionWeightPlotter:

    def __init__(self, tokenizer):

        self.tokenizer = tokenizer

    def __call__(self, attention, sentence, result, layer):
        fig = plt.figure(figsize=(16, 8))

        sentence = self.tokenizer.encode(sentence)

        attention = tf.squeeze(attention[layer], axis=0)

        for head in range(attention.shape[0]):
            ax = fig.add_subplot(2, 4, head + 1)

            # plot the attention weights
            ax.matshow(attention[head][:-1, :], cmap='viridis')

            fontdict = {'fontsize': 10}

            ax.set_xticks(range(len(sentence) + 2))
            ax.set_yticks(range(len(result)))

            ax.set_ylim(len(result) - 1.5, -0.5)

            ax.set_xticklabels(
                ['<start>'] + [self.tokenizer.decode([i]) for i in sentence] + ['<end>'],
                fontdict=fontdict, rotation=90)

            ax.set_yticklabels([self.tokenizer.decode([i]) for i in result
                                if i < self.tokenizer.vocab_size],
                               fontdict=fontdict)

            ax.set_xlabel('Head {}'.format(head + 1))

        plt.tight_layout()
        plt.show()


class Chatbot:
    # Note : only deterministic output (depends on evaluator)

    def __init__(self, evaluator):

        self.evaluator = evaluator
        self.tokenizer = self.evaluator.tokenizer

        self.attention_weights_plotter = AttentionWeightPlotter(self.tokenizer)

    def __call__(self, sentence, layer_attention_to_plot_idx=None):
        result, attention_weights = self.evaluator(sentence)

        predicted_sentence = self.tokenizer.decode([i for i in result
                                                    if i < self.tokenizer.vocab_size])

        print('User input: {}'.format(sentence))
        print('Bot output: {}'.format(predicted_sentence))

        if layer_attention_to_plot_idx is not None:
            self.attention_weights_plotter(attention_weights,
                                           sentence,
                                           result,
                                           layer_attention_to_plot_idx)



import traceback

import abc
import os

from mlpug.base import Base

from mlpug.mlpug_logging import get_logger

DEBUG_MULTI_PROCESSING = os.environ['DEBUG_MULTI_PROCESSING'] if 'DEBUG_MULTI_PROCESSING' in os.environ else False


class MultiProcessingContextBase(Base, metaclass=abc.ABCMeta):

    def __init__(self, name="MultiProcessingContext", **kwargs):
        super().__init__(pybase_logger_name=name, **kwargs)

    @abc.abstractmethod
    def is_distributed(self):
        raise NotImplemented("Please implement this method in your child class")

    @abc.abstractmethod
    def is_primary(self):
        raise NotImplemented("Please implement this method in your child class")

    @abc.abstractmethod
    def device_rank(self):
        raise NotImplemented("Please implement this method in your child class")

    @abc.abstractmethod
    def world_size(self):
        raise NotImplemented("Please implement this method in your child class")


class MultiProcessingManager:

    _log = get_logger("MultiProcessingManager")

    _context = None

    @classmethod
    def set_context(cls, context):
        cls._context = context

        if bool(DEBUG_MULTI_PROCESSING) is True:
            cls._log.info(f"Using multi-processing context {str(context)}.")

    @classmethod
    def get_context(cls):
        return cls._context


class MultiProcessingMixin:

    def __init__(self, *args,
                 is_distributed=None,
                 is_primary=None,
                 device_rank=None,
                 disable_logging=None,
                 **kwargs):

        mp_context = MultiProcessingManager.get_context()

        if is_distributed is None:
            is_distributed = mp_context.is_distributed()

        if device_rank is None:
            device_rank = mp_context.device_rank() if is_distributed else 0

        if is_primary is None:
            is_primary = (not is_distributed) or mp_context.is_primary()

        if is_distributed and bool(DEBUG_MULTI_PROCESSING) is True:
            disable_logging = False

        if disable_logging is None:
            disable_logging = is_distributed and not is_primary

        self._is_distributed = is_distributed
        self._device_rank = device_rank
        self._is_primary = is_primary

        super().__init__(*args, disable_logging=disable_logging, **kwargs)

    @property
    def is_distributed(self):
        return self._is_distributed

    @property
    def is_primary(self):
        return self._is_primary

    @property
    def device_rank(self):
        return self._device_rank

    def _pybase_get_logger_name(self):
        if self.is_distributed:
            return f"[Device {self.device_rank}] {super()._pybase_get_logger_name()}"

        return super()._pybase_get_logger_name()



from mlpug.trainers.callbacks.metrics_logger import *

from mlpug.trainers.callbacks.metrics_logger import \
    TrainingMetricsLogger as TrainingMetricsLoggerBase, \
    TestMetricsLogger as TestMetricsLoggerBase

from mlpug.pytorch.multi_processing import MultiProcessingMixin

from mlpug.pytorch.utils import SlidingWindow


class TrainingMetricsLogger(MultiProcessingMixin, TrainingMetricsLoggerBase):

    def __init__(self, *args, sliding_window_factory=SlidingWindow, **kwargs):
        super().__init__(*args, sliding_window_factory=sliding_window_factory, **kwargs)


class TestMetricsLogger(MultiProcessingMixin, TestMetricsLoggerBase):

    def __init__(self, *args, sliding_window_factory=SlidingWindow, **kwargs):
        super().__init__(*args, sliding_window_factory=sliding_window_factory, **kwargs)



import os
import argparse

import time

import random

import pickle

import numpy as np

from basics.logging import get_logger

logger = get_logger(os.path.basename(__file__))

parser = argparse.ArgumentParser(description='Split the given dataset in to multiple subsets, as defined by '
                                             '--split and --set-names')

parser.add_argument(
    '--conversations-dataset',
    type=str, required=True,
    help='path to the conversations dataset')

parser.add_argument(
    '--split',
    type=float, nargs='+', required=False, default=[0.7, 0.3],
    help='Names of sets to split the original set in to')

parser.add_argument(
    '--set-names',
    type=str, nargs='+', required=False, default=['training', 'validation'],
    help='Names of sets to split the original set in to')


args = parser.parse_args()


def load_conversations_data(conversations_dataset_file):
    logger.info(f'\nLoading conversations dataset file : {conversations_dataset_file}')

    with open(conversations_dataset_file, 'rb') as f:
        data = pickle.load(f)
        return data


def save_conversations_data(conversations_dataset_file, pairs, original_metadata, voc=None):
    logger.info(f'\nSaving conversations dataset file : {conversations_dataset_file}')
    logger.info(f'\nData set size : {len(pairs)}')

    with open(conversations_dataset_file, 'wb') as f:

        conversation_data = {
            'pairs': pairs,
            'args': args,
            'script': os.path.basename(__file__),
            'original_metadata': original_metadata
        }

        if voc is not None:
            conversation_data['voc'] = voc

        pickle.dump(conversation_data, f)


def print_samples(pairs):
    # Print some conversations to validate
    for pair in pairs[:10]:
        logger.info(f"\n{pair[0]}\n{pair[1]}\n")


split = np.array(args.split)
logger.info(f'split : {split}')
if split.sum() != 1.0:
    logger.error("The split ratios should sum up to one")
    exit(-1)

set_names = args.set_names
logger.info(f'set_names : {set_names}')
if len(split) != len(set_names):
    logger.error("The number or set names given should be equal to the number of split ratios given")
    exit(-2)

conversations_dataset_file = args.conversations_dataset
data = load_conversations_data(conversations_dataset_file)

pairs = data['pairs']
voc = data['voc'] if 'voc' in data else None

original_dataset_metadata = data.copy()
del original_dataset_metadata['pairs']

if voc is not None:
    del original_dataset_metadata['voc']

# TODO : This code should be DRYed up
num_conversations = len(pairs)
logger.info(f'Number of conversations in dataset : {num_conversations}')
indices = list(range(num_conversations))
random.shuffle(indices)

set_lengths = list(map(lambda r: int(r*num_conversations), split))
rest = num_conversations - sum(set_lengths)
set_lengths[-1] += rest

if sum(set_lengths) != num_conversations:
    logger.error("Unexpected : length of subsets is not equal to given dataset")
    exit(-3)

data_path, base_filename = os.path.split(conversations_dataset_file)

current_time = int(time.time())

start_idx = 0
for name, l in zip(set_names, set_lengths):
    logger.info(f'Creating subset {name} of size {l} ...')
    dataset_filename = f'{name}-{current_time}-{base_filename}'
    dataset_path = os.path.join(data_path, dataset_filename)

    indices_subset = indices[start_idx:start_idx+l]

    conversations_subset = [pairs[i] for i in indices_subset]
    print_samples(conversations_subset)

    save_conversations_data(dataset_path, conversations_subset, original_dataset_metadata, voc)

    start_idx += l



from mlpug.pytorch.trainers.callbacks.basic import *
from mlpug.pytorch.trainers.callbacks.callback import *
from .checkpoint_manager import CheckpointManager
from mlpug.pytorch.trainers.callbacks.lr_scheduler_wrapper import LRSchedulerWrapper
from mlpug.pytorch.trainers.callbacks.metrics_logger import MetricsLoggingMode, TrainingMetricsLogger, TestMetricsLogger
from mlpug.pytorch.trainers.callbacks.tensorboard import Tensorboard, AutoTensorboard
from mlpug.pytorch.trainers.callbacks.distributed import DistributedSamplerManager
from mlpug.pytorch.trainers.callbacks.cuda_memory import EmptyCudaCache, LogCudaMemory



import os
import itertools

import torch

from basics.logging import get_logger

from mlpug.examples.chatbot.conversation_dataset import indexesFromSentence

logger = get_logger(os.path.basename(__file__))


def zeroPadding(l, fillvalue):
    return list(itertools.zip_longest(*l, fillvalue=fillvalue))


def binaryMatrix(l, value):
    m = []
    for i, seq in enumerate(l):
        m.append([])
        for token in seq:
            if token == value:
                m[i].append(0)
            else:
                m[i].append(1)
    return m


# Returns padded input sequence tensor and lengths
def inputVar(sentences, voc, PAD_token, EOS_token):
    indexes_batch = [indexesFromSentence(voc, sentence, EOS_token) for sentence in sentences]
    lengths = torch.tensor([len(indexes) for indexes in indexes_batch])
    # Batch dimension should be second (in order to partition over multiple GPUs)
    lengths = lengths.unsqueeze(0)

    padList = zeroPadding(indexes_batch, PAD_token)
    padVar = torch.LongTensor(padList)

    return padVar, lengths


# Returns padded target sequence tensor, padding mask, and max target length
def outputVar(sentences, voc, PAD_token, EOS_token):
    indexes_batch = [indexesFromSentence(voc, sentence, EOS_token) for sentence in sentences]
    max_target_len = max([len(indexes) for indexes in indexes_batch])
    padList = zeroPadding(indexes_batch, PAD_token)
    mask = binaryMatrix(padList, PAD_token)
    mask = torch.BoolTensor(mask)
    padVar = torch.LongTensor(padList)
    return padVar, mask, max_target_len


# Returns all items for a given batch of pairs
def batch2TrainData(voc, pair_batch, PAD_token, EOS_token):
    pair_batch.sort(key=lambda x: len(x[0].split(" ")), reverse=True)
    input_batch, output_batch = [], []
    for pair in pair_batch:
        input_batch.append(pair[0])
        output_batch.append(pair[1])
    inp, lengths = inputVar(input_batch, voc, PAD_token, EOS_token)
    output, mask, max_target_len = outputVar(output_batch, voc, PAD_token, EOS_token)
    return inp, lengths, output, mask, max_target_len


def create_sentence_pairs_collate_fn(PAD_token, fixed_sequence_length=None):
    """

    `fixed_sequence_length` is important when training on TPU

    :param PAD_token:
    :type PAD_token:
    :param fixed_sequence_length:
    :type fixed_sequence_length:
    :return:
    :rtype:
    """

    if fixed_sequence_length:
        logger.info(f"Using fixed sequence lengths of {fixed_sequence_length} tokens.")

    def collate_fn(indexed_sentence_pairs):
        # Why is the sort required?
        # ==> This is a CuDNN requirement
        # ==> https://discuss.pytorch.org/t/why-lengths-should-be-given-in-sorted-order-in-pack-padded-sequence/3540
        # ==> Apparently solved now?
        indexed_sentence_pairs.sort(key=lambda pair: len(pair[0]), reverse=True)

        input_batch, output_batch = [], []
        for pair in indexed_sentence_pairs:
            input_batch.append(pair[0])
            output_batch.append(pair[1])

        # ############# PROCESS INPUT BATCH #############
        input_lengths = torch.tensor([len(indexed_sentence) for indexed_sentence in input_batch], dtype=torch.short)
        # Batch dimension should be second (in order to partition over multiple GPUs)
        input_lengths = input_lengths.unsqueeze(0)

        if fixed_sequence_length:
            padded_input_batch = torch.ones(fixed_sequence_length, len(input_batch), dtype=torch.long) * PAD_token
            for idx, indexed_sentence in enumerate(input_batch):
                padded_input_batch[0:len(indexed_sentence), idx] = torch.LongTensor(indexed_sentence)
        else:
            padded_input_batch = zeroPadding(input_batch, PAD_token)
            padded_input_batch = torch.LongTensor(padded_input_batch)

        ################################################

        # ############# PROCESS OUTPUT BATCH ############
        if fixed_sequence_length:
            max_output_len = fixed_sequence_length
            padded_output_batch = torch.ones(fixed_sequence_length, len(output_batch), dtype=torch.long) * PAD_token
            for idx, indexed_sentence in enumerate(output_batch):
                padded_output_batch[0:len(indexed_sentence), idx] = torch.LongTensor(indexed_sentence)
            output_mask = padded_output_batch != PAD_token
        else:
            max_output_len = max([len(indexed_sentence) for indexed_sentence in output_batch])
            padded_output_batch = zeroPadding(output_batch, PAD_token)

            output_mask = binaryMatrix(padded_output_batch, PAD_token)
            output_mask = torch.BoolTensor(output_mask)

            padded_output_batch = torch.LongTensor(padded_output_batch)
        ################################################

        return padded_input_batch, input_lengths, padded_output_batch, output_mask, max_output_len

    return collate_fn



import io

import tensorflow as tf
import h5py
from tensorflow.python.keras.saving import hdf5_format

import basics.base_utils as _

from mlpug.trainers.training import *

from mlpug.mlpug_exceptions import TrainerInvalidException, \
    TrainerStateInvalidException, \
    BatchNotChunkableException, \
    MLPugException, \
    LossNotAvailableException

from mlpug.utils import get_value_at


class TFTrainerMixin:

    def _activate_inference_mode(self, inference_mode):
        # No pre-evaluation mode change
        pass

    def _get_model_state(self, model, model_name=None):
        state = io.BytesIO()
        with h5py.File(state, 'w') as f:
            hdf5_format.save_weights_to_hdf5_group(f, model.layers)

        return state

    def _get_optimizer_state(self, optimizer, optimizer_name=None):
        state = io.BytesIO()
        with h5py.File(state, 'w') as f:
            hdf5_format.save_optimizer_weights_to_hdf5_group(f, optimizer)

        return state

    def _set_model_state(self, model, state, model_name=None):
        with h5py.File(state, 'r') as f:
            hdf5_format.load_weights_from_hdf5_group(f, model.layers)

    def _set_optimizer_state(self, optimizer, state, optimizer_name):
        with h5py.File(state, 'r') as f:
            weights = hdf5_format.load_optimizer_weights_from_hdf5_group(f)
            optimizer.set_weights(weights)


class Trainer(TFTrainerMixin, TrainerBase):
    pass


class DefaultTrainer(TFTrainerMixin, DefaultTrainerBase):

    def __init__(self, *args,
                 eager_mode=False,
                 batch_data_signature=None,
                 training_settings_signature=None,
                 distribution_strategy=None,
                 trainable_variables=None,
                 name="DefaultTrainer",
                 **kwargs):
        """

        :param args:
        :param eager_mode: If true, the training step is not wrapped in a @tf.function
        :param batch_data_signature: Is only required when eager_mode=False
                                     Example, when batch data is a tuple of an input and target tensor
                                     (tf.TensorSpec(shape=(None, None), dtype=tf.int64),
                                      tf.TensorSpec(shape=(None, None), dtype=tf.int64),)

        :param training_settings_signature: Use when you use training_settings.
                                            Is only used when eager_mode=False

                                            Default is {}.

                                            Note: training_settings, are the same as evaluation_settings.

        :param distribution_strategy: Optional distributed training strategy

        :param trainable_variables: Only required when using multiple optimizers

        :param kwargs:
        """
        super(DefaultTrainer, self).__init__(*args, **kwargs)

        self.eager_mode = eager_mode
        self.batch_data_signature = batch_data_signature
        self.training_settings_signature = training_settings_signature

        self.distribution_strategy = distribution_strategy

        self.trainable_variables = trainable_variables

        if not eager_mode:
            if self.batch_data_signature is None:
                raise TrainerInvalidException(f"Missing batch_data_signature such that the "
                                              f"training step computation graph can be traced")

            if self.training_settings_signature is None:
                self._log.info("training_settings_signature not given, setting to empty dict, "
                               "implying that training settings won't be used.")
                self.training_settings_signature = {}

            self._train_step_tf_func = self._create_training_step_tf_func() if self.distribution_strategy is None \
                else self._create_distributed_training_step_tf_func()

            self._call_model_tf_func = self._create_call_model_tf_func() if self.distribution_strategy is None \
                else self._create_distributed_call_model_tf_func()
        else:
            self._log.warn("Training in eager mode.")
            self._train_step_tf_func = self._train_on if self.distribution_strategy is None \
                else self._create_distributed_training_step_eager()

            self._call_model_tf_func = self._call_model if self.distribution_strategy is None \
                else self._create_distributed_call_model_eager()

        if self.trainable_variables is None:
            if len(self.optimizers) > 1:
                raise TrainerInvalidException(f"No trainable variables provided per optimizer")
        else:
            self.trainable_variables = convert_to_dict("optimizer", trainable_variables)

            missing_optimizer_vars = []
            for optimizer_name in self.optimizers.keys():
                if optimizer_name not in self.trainable_variables or self.trainable_variables[optimizer_name] is None:
                    missing_optimizer_vars += [optimizer_name]

            if len(missing_optimizer_vars) > 0:
                raise TrainerInvalidException(f"Missing trainable variables for optimizer(s) : "
                                              f"{', '.join(missing_optimizer_vars)}")

        self._deferred_model_components_state = None
        self._deferred_optimizers_state = None

        self._first_batch = True

    def set_model_components_state(self, state):
        """

        :param state:
        :return: success (True or False)
        """
        if not _.is_callable(getattr(state, 'items', None)):
            self._log.error("State is invalid, unable to set model components state")
            return False

        self._deferred_model_components_state = state

        self._log.debug("Model components checkpoint state received; "
                        "deferred setting the state until training has started")

        return True

    def set_optimizers_state(self, state):
        """

        :param state:
        :return: success (True, False)
        """
        if not _.is_callable(getattr(state, 'items', None)):
            self._log.error("State is invalid, unable to set optimizers state")
            return False

        self._deferred_optimizers_state = state

        self._log.debug("Optimizers checkpoint state received; "
                        "deferred setting the state until training has started")

        return True

    def set_learning_rate_for(self, optimizer_name, lr):
        """

        Set learning rate for specific optimizer `optimizer_name` to `lr`

        :param optimizer_name:
        :param lr:

        :return: True on success, else False
        """
        optimizer = self.get_optimizer(optimizer_name)
        if not hasattr(optimizer, 'learning_rate'):
            self._log.error(f"No valid optimizer available with name {optimizer_name}, unable to set learning rate")
            return False

        try:
            optimizer.learning_rate = lr
        except Exception as e:
            _.log_exception(self._log, f"Unable to set learning rate for optimizer {optimizer_name}", e)
            return False

        self._log.debug(f"Learning rate of optimizer {optimizer_name} set to : {lr}")

        return True

    def train_on(self, batch_data, training_settings=None):
        """
        Use batch_data to perform a training iteration.

        Optionally uses `batch_chunk_size` to evaluate the loss in chunks.
        If a `batch_chunk_size` was given during construction of the trainer, the gradients are updated by evaluating
        the batch in chunks.

        *Note*
        When using chunked batch processing, the default implementation assumes that the
        loss, calculated over a chunk, is the average of the sample losses.

        :param batch_data: batch_data object to train on (e.g. dict, list, tuple)
                           When `batch_chunk_size` is given, `batch_data` must be an object that implements the
                           `__len__` and `__getitem__` methods. Here the `__getitem__` method must be able to deal
                           with slices.
        :param training_settings: optional training_settings object (usually dict)

        :return: loss, auxiliary_results

        loss : number (e.g. float)
        auxiliary_results : can be anything, e.g dict or list with values or data items
        """

        if not self.instance_valid():
            raise TrainerInvalidException()

        if self._first_batch:
            # Check if we first need to restore a checkpoint
            deferred_model_components_state_set = \
                self._set_deferred_model_components_state(batch_data, training_settings)

            if deferred_model_components_state_set:
                # To set the deferred_model_components_state at the first batch the model was evaluated
                # So we can get the trainable variables from the model and subsequently set the
                # deferred optimizer state, which needs teh trainable variables.
                self._retrieve_trainable_variables()
                self._set_deferred_optimizers_state()

            self._first_batch = False

        loss, auxiliary_results = self._train_step_tf_func(batch_data, training_settings)

        return loss, auxiliary_results

    def _create_train_step_signature(self):
        return [
            self.batch_data_signature,
            self.training_settings_signature
        ]

    def _create_distributed_training_step_tf_func(self):
        @tf.function(input_signature=self._create_train_step_signature())
        def training_step_func(batch_data, training_settings):
            return self.distribution_strategy.run(self._train_on, args=(batch_data, training_settings,))

        return training_step_func

    def _create_training_step_tf_func(self):
        @tf.function(input_signature=self._create_train_step_signature())
        def training_step_func(batch_data, training_settings):
            return self._train_on(batch_data, training_settings)

        return training_step_func

    def _create_distributed_training_step_eager(self):
        def training_step_func(batch_data, training_settings):
            return self.distribution_strategy.run(self._train_on, args=(batch_data, training_settings,))

        return training_step_func

    def _train_on(self, batch_data, training_settings):
        loss, auxiliary_results, gradients = self._calc_gradients(batch_data, training_settings=training_settings)

        self._update_model_parameters(self._prepare_update_model_parameters(gradients))

        self._after_update_model_parameters(gradients)

        return loss, auxiliary_results

    def _create_call_model_signature(self):
        return [
            self.batch_data_signature,              # batch_data
            self.training_settings_signature,       # evaluate_settings
            tf.TensorSpec(shape=(), dtype=tf.bool)  # inference_mode
        ]

    def _create_distributed_call_model_tf_func(self):
        @tf.function(input_signature=self._create_call_model_signature())
        def call_model_func(batch_data, evaluate_settings, inference_mode):
            return self.distribution_strategy.run(self._call_model,
                                                  args=(batch_data, evaluate_settings, inference_mode))

        return call_model_func

    def _create_call_model_tf_func(self):
        @tf.function(input_signature=self._create_call_model_signature())
        def call_model_func(batch_data, evaluate_settings, inference_mode):
            return self._call_model(batch_data, evaluate_settings, inference_mode)

        return call_model_func

    def _create_distributed_call_model_eager(self):
        def call_model_func(batch_data, evaluate_settings, inference_mode):
            return self.distribution_strategy.run(self._call_model,
                                                  args=(batch_data, evaluate_settings, inference_mode))

        return call_model_func

    def _retrieve_trainable_variables(self):
        if len(self.optimizers) > 1:
            return

        # This only needs to be done once
        # Further, this situation only occurs when there is only one optimizer

        optimizer_name = next(iter(self.optimizers))
        trainable_variables = get_value_at(optimizer_name, self.trainable_variables, warn_on_failure=False)
        if trainable_variables is None:
            trainable_variables = self.training_model.trainable_variables

            self.trainable_variables = {
                optimizer_name: trainable_variables
            }

    def _set_deferred_model_components_state(self, batch_data, training_settings):
        """
        Model component state can only be set after evaluating the model on input data
        (Crazy but true)

        :param batch_data:
        :param training_settings:
        :return: True if set, else False
        """

        if self._deferred_model_components_state is None:
            return False

        def dry_eval_model():
            self.evaluate_loss(batch_data,
                               inference_mode=False,
                               evaluate_settings=training_settings)

        if self.distribution_strategy is not None:
            with self.distribution_strategy.scope():
                dry_eval_model()
        else:
            dry_eval_model()

        success = super().set_model_components_state(self._deferred_model_components_state)
        if not success:
            self._log.error("Unable to set deferred model components state, weights are not loaded")

        self._deferred_model_components_state = None

        return success

    def _set_deferred_optimizers_state(self):
        if self._deferred_optimizers_state is None:
            return

        def create_optimizer_weights():
            for optimizer_name, optimizer in self.optimizers.items():
                trainable_variables = get_value_at(optimizer_name, self.trainable_variables, warn_on_failure=False)
                optimizer._create_all_weights(trainable_variables)

        if self.distribution_strategy is not None:
            with self.distribution_strategy.scope():
                create_optimizer_weights()
        else:
            create_optimizer_weights()

        success = super().set_optimizers_state(self._deferred_optimizers_state)
        if not success:
            self._log.error("Unable to set deferred optimizers state, weights are not loaded")

        self._deferred_optimizers_state = None

    def _evaluate_loss(self, batch_data, evaluate_settings=None, inference_mode=None):
        """
        Evaluates the given training model on the  given batch_data, using the optional training_settings
        Depending on the Deep learning backend you might need to use inference mode here

        :param batch_data: batch_data object to evaluate loss on (e.g. dict, list, tuple)
        :param evaluate_settings: optional evaluate_settings object (usually dict)
        :param inference_mode: optional bool, important when inference mode not set in `_activate_inference_mode`
                               Pytorch:     inference_mode not required here
                               Tensorflow:  inference_mode required here

        :return: dict or tuple
            {
                "loss": <Tensor>,
                "auxiliary_results": <can be anything, e.g dict or list with values or data items>
            }

            (loss, ... auxiliary results ...)
        """

        if not (type(inference_mode) is bool):
            raise TrainerStateInvalidException("Inference mode is not set")

        if not inference_mode:
            # @tf.function on the train step level
            return self._call_model(batch_data, evaluate_settings, inference_mode)
        else:
            return self._call_model_tf_func(batch_data, evaluate_settings, tf.constant(inference_mode, dtype=tf.bool))

    def _call_model(self, batch_data, evaluate_settings, inference_mode):
        return self.training_model(batch_data, evaluate_settings, inference_mode)

    def _calc_gradients(self, batch_data, training_settings=None):
        """

        :param batch_data:
        :param training_settings:
        :return:

        :raises LossNotAvailableException
        """

        if not self.batch_chunk_size:
            with tf.GradientTape() as tape:
                results = self.evaluate_loss(batch_data,
                                             inference_mode=False,
                                             evaluate_settings=training_settings)

            if 'loss' not in results:
                raise LossNotAvailableException()

            if self.trainable_variables is None:
                # We now have evaluated the model and the trainable variables should be available
                self._retrieve_trainable_variables()

            loss = results['loss']
            auxiliary_results = get_value_at('auxiliary_results', results, warn_on_failure=False)

            gradients = self._back_propagate_from(loss, tape)
        else:
            raise NotImplementedError("Gradient accumulation over batch chunks is not implemented")

        return loss, auxiliary_results, gradients

    def _back_propagate_from(self, loss, tape, last_chunk=False):
        gradients = {}
        for optimizer_name in self.optimizers.keys():
            trainable_variables = get_value_at(optimizer_name, self.trainable_variables, warn_on_failure=False)

            gradients[optimizer_name] = tape.gradient(loss, trainable_variables)

        return gradients

    def _prepare_update_model_parameters(self, gradients):
        """

        :param gradients: dict with gradients per provided optimizer
                          The simple situation, when only one optimizer is given, the structure would be:
                          {
                              'optimizer': <gradients>
                          }
        :return: processed gradients with the same dict structure
        """
        return gradients

    def _update_model_parameters(self, gradients):
        for optimizer_name, optimizer in self.get_optimizers().items():
            trainable_variables = get_value_at(optimizer_name, self.trainable_variables)
            if trainable_variables is None:
                raise MLPugException("Unexpected state :  trainable variables not found. Please file an issue.")

            optimizer.apply_gradients(zip(gradients[optimizer_name], trainable_variables))

    def _after_update_model_parameters(self, gradients):
        pass



import logging

ESCAPE = '\033'
FG_COLOR_CODE = '[38;5;%dm'
BG_COLOR_CODE = '[48;5;%dm'
RESET_CODE = '[0;00m'


class FormatterBase(logging.Formatter):

    @staticmethod
    def apply_fmt(format_code, level, s, color_map):
        color = color_map[level] if level in color_map else None
        if color is None:
            return s

        return ESCAPE + (format_code % color) + s

    @staticmethod
    def finalize_fmt(s):
        return s + ESCAPE + RESET_CODE



"""
Conversation dataset class and utils
"""


# From https://pytorch.org/tutorials/beginner/chatbot_tutorial.html
# but not PyTorch specific
import pickle


def indexesFromSentence(voc, sentence, EOS_token):
    return [voc.word2index[word] for word in sentence.split(' ')] + [EOS_token]


class IndexedSentencePairsDataset:
    def __init__(self, sentence_pairs, voc, EOS_token):
        super(IndexedSentencePairsDataset).__init__()

        self.sentence_pairs = sentence_pairs
        self.voc = voc

        self.EOS_token = EOS_token

    def __len__(self):
        return len(self.sentence_pairs)

    def __getitem__(self, idx):
        input_sentence = self.sentence_pairs[idx][0]
        output_sentence = self.sentence_pairs[idx][1]

        return self._map_to_indices(input_sentence), self._map_to_indices(output_sentence)

    def _map_to_indices(self, sentence):
        return indexesFromSentence(self.voc, sentence, self.EOS_token)


def load_sentence_pair_data(sentence_pair_file, logger=None):
    if logger is not None:
        logger.info(f'Loading sentence pair file : {sentence_pair_file}\n')

    with open(sentence_pair_file, 'rb') as f:
        data = pickle.load(f)
        return data['pairs'], data['voc']


import torch_xla.core.xla_model as xm

import numpy as np

from mlpug.trainers.training import BatchChunkingResults
from mlpug.evaluation import default_metric_reducer_func

from mlpug.pytorch.evaluation import MetricEvaluator as MetricEvaluatorPyTorch

from mlpug.utils import is_chunkable

from mlpug.pytorch.evaluation import \
    GatherLossSimple as GatherLossSimplePytorch, \
    GatherMaskedLoss as GatherMaskedLossPyTorch


class GatherLossSimple(GatherLossSimplePytorch):

    def _gather_loss_distributed(self, loss, **kwargs):
        loss_sum = xm.mesh_reduce('gather_loss', loss, np.sum)
        num_devices = xm.xrt_world_size()
        loss = loss_sum / num_devices

        return loss, loss_sum, num_devices


class GatherMaskedLoss(GatherMaskedLossPyTorch):

    def _gather_loss_distributed(self, loss_sum, num_samples, **kwargs):
        loss_sum = xm.mesh_reduce('gather_loss', loss_sum, np.sum)
        num_samples = xm.mesh_reduce('gather_num_samples', num_samples, np.sum)

        loss = loss_sum / num_samples

        return loss, loss_sum, num_samples


class MetricEvaluator(MetricEvaluatorPyTorch):

    def __init__(self, *args, batch_metric_funcs=None, name="MetricEvaluator", **kwargs):

        if batch_metric_funcs is None:
            batch_metric_funcs = {
                "loss": GatherLossSimple(requester=name)
            }

        super().__init__(*args, batch_metric_funcs=batch_metric_funcs, name=name, **kwargs)



import os
import sys
import random
import time

import torch
import torch.nn as nn

from mlpug.examples.chatbot.pytorch.original_chatbot_tutorial.loss import cross_entropy, masked_loss
from mlpug.examples.chatbot.pytorch.original_chatbot_tutorial.model_data_generation import batch2TrainData

# NVIDIA Automatic Mixed Precision Module
# Only imported when needed
amp = None


class Seq2SeqTrainModel(nn.Module):
    """
    Use this module to make training fast when using nn.DataParallel()
    """
    def __init__(self, encoder, decoder):
        super(Seq2SeqTrainModel, self).__init__()

        self.encoder = encoder
        self.decoder = decoder

    def forward(self,
                chat_inputs,
                seq_lengths,
                init_decoder_input,
                max_output_len,
                expected_bot_outputs,
                use_teacher_forcing):
        batch_size = chat_inputs.size(1)

        # Forward pass through encoder
        encoder_outputs, encoder_hidden = self.encoder(chat_inputs, seq_lengths)

        decoder_n_layers = self.decoder.n_layers
        # Set initial decoder hidden state to the encoder's final hidden state
        # This is incorrect:
        # decoder_hidden = encoder_hidden[:decoder_n_layers]

        # Taken the average over both directions
        encoder_hidden = encoder_hidden.view(self.encoder.n_layers,
                                             2,
                                             -1,
                                             self.encoder.hidden_size)

        decoder_hidden = 0.5*(encoder_hidden[:decoder_n_layers, 0, :, :] + encoder_hidden[:decoder_n_layers, 1, :, :])

        decoder_input = init_decoder_input

        # What happens to this, will this be available directly on the right device?
        per_sample_cross_entropy = torch.FloatTensor(max_output_len, batch_size).to(encoder_outputs.device)
        # Forward batch of sequences through decoder one time step at a time
        for t in range(max_output_len):
            decoder_output, decoder_hidden = self.decoder(decoder_input, decoder_hidden, encoder_outputs)

            if use_teacher_forcing:
                decoder_input = expected_bot_outputs[t]
                decoder_input = decoder_input.unsqueeze(0)
            else:
                # No teacher forcing: next input is decoder's own current output
                _, top_1_index = decoder_output.topk(1)
                decoder_input = torch.LongTensor([[top_1_index[i][0] for i in range(batch_size)]])

            decoder_output = decoder_output.squeeze()
            per_sample_cross_entropy[t, :] = cross_entropy(decoder_output, expected_bot_outputs[t])

        return {
            "loss": per_sample_cross_entropy
        }


def train(input_variable, lengths, init_decoder_input, max_output_len, target_variable, mask, teacher_forcing_ratio,
          train_model, optimizer, use_mixed_precision, clip):

    # mixed precision
    global amp

    # Zero gradients
    optimizer.zero_grad()

    # Initialize variables

    # Determine if we are using teacher forcing this iteration
    use_teacher_forcing = True if random.random() < teacher_forcing_ratio else False

    results = train_model(input_variable, lengths, init_decoder_input, max_output_len,
                         target_variable, use_teacher_forcing)

    per_sample_loss = results["loss"]

    loss = masked_loss(per_sample_loss, mask)

    # Perform back propagation
    if use_mixed_precision:
        if not amp:
            from apex import amp

        with amp.scale_loss(loss, optimizer) as scaled_loss:
            scaled_loss.backward()
    else:
        loss.backward()

    # Clip gradients: gradients are modified in place
    _ = nn.utils.clip_grad_norm_(train_model.parameters(), clip)

    # Adjust model weights
    optimizer.step()

    return loss.item()


def trainIters(model_name, voc, pairs, PAD_token, EOS_token, SOS_token, train_model, optimizer,
               model_checkpoint_dir, n_iteration, batch_size, teacher_forcing_ratio, use_mixed_precision, clip,
               print_every, save_every, corpus_name, checkpoint, device):

    # Was DataParallel applied?
    if hasattr(train_model, 'encoder'):
        embedding = train_model.encoder.embedding
        encoder = train_model.encoder
        decoder = train_model.decoder
    else:
        module = train_model.module
        embedding = module.encoder.embedding
        encoder = module.encoder
        decoder = module.decoder

    embedding_dim = embedding.embedding_dim
    encoder_state_size = encoder.hidden_size
    encoder_n_layers = encoder.n_layers
    decoder_n_layers = decoder.n_layers

    # Load batches for each iteration
    print("Generating data batches ...")
    training_batches = [batch2TrainData(voc, [random.choice(pairs) for _ in range(batch_size)], PAD_token, EOS_token)
                        for _ in range(n_iteration)]

    # Initializations
    print('Initializing ...')
    start_iteration = 1
    print_loss = 0
    if checkpoint and 'iteration' in checkpoint:
        start_iteration = checkpoint['iteration'] + 1

    init_decoder_input = torch.tensor([[SOS_token for _ in range(batch_size)]], dtype=torch.long, device=device)

    # Training loop
    print("Training...")
    tot_iter_time = 0
    for iteration in range(start_iteration, n_iteration + 1):
        start_time = time.time()

        training_batch = training_batches[iteration - 1]
        # Extract fields from batch
        input_variable, lengths, target_variable, mask, max_target_len = training_batch

        # Set device options
        input_variable = input_variable.to(device)
        lengths = lengths.to(device)
        target_variable = target_variable.to(device)
        mask = mask.to(device)

        # Run a training iteration with batch
        loss = train(input_variable, lengths, init_decoder_input, max_target_len, target_variable, mask, teacher_forcing_ratio,
                     train_model, optimizer, use_mixed_precision, clip)
        print_loss += loss

        end_time = time.time()
        iter_time = end_time-start_time
        tot_iter_time += iter_time

        # Print progress
        if iteration % print_every == 0:
            print_loss_avg = print_loss / print_every
            print("Iteration: {}; Percent complete: {:.1f}%; Average loss: {:.4f}; Average iter. duration: {}ms".format(
                iteration,
                iteration / n_iteration * 100,
                print_loss_avg,
                int(1000*tot_iter_time/print_every)))
            tot_iter_time = 0
            print_loss = 0

        # Save checkpoint
        if iteration % save_every == 0 or iteration == n_iteration:
            print("Saving checkpoint ... ")
            directory = os.path.join(model_checkpoint_dir, model_name, corpus_name, '{}-{}-{}-{}'.format(
                embedding_dim,
                encoder_n_layers,
                decoder_n_layers,
                encoder_state_size))
            if not os.path.exists(directory):
                os.makedirs(directory)
            torch.save({
                'iteration': iteration,
                'en': encoder.state_dict(),
                'de': decoder.state_dict(),
                'opt': optimizer.state_dict(),
                'loss': loss,
                'voc_dict': voc.__dict__,
                'embedding': embedding.state_dict()
            }, os.path.join(directory, '{}_{}.tar'.format(iteration, 'checkpoint')))

        sys.stdout.flush()



import os

import pickle

import tensorflow as tf
import tensorflow_datasets as tfds

from mlpug.examples.chatbot.shared import create_argument_parser
from mlpug.examples.chatbot.conversation_dataset import load_sentence_pair_data

from mlpug.examples.chatbot.tensorflow.original_transformer_tutorial.model_data_generation import \
    create_chatbot_tf_encode_func, \
    create_length_filter_func

from mlpug.examples.chatbot.tensorflow.original_transformer_tutorial.transformer import Transformer

from mlpug.examples.chatbot.tensorflow.original_transformer_tutorial.training import TrainModel, CustomSchedule

import mlpug.tensorflow as mlp


from basics.logging import get_logger

BUFFER_SIZE = 1000


def dataset_path_for(subset):
    return os.path.join(dataset_path, f"{subset}-{base_dataset_filename}")


def create_dataset_generator(pairs):

    def generator():
        for pair in pairs:
            yield tuple(pair)

    return generator


def prepare_dataset(raw_dataset, tf_encode, filter_sequence_length):
    dataset = tf.data.Dataset.from_generator(
        create_dataset_generator(raw_dataset),
        (tf.string, tf.string),
        (tf.TensorShape([]), tf.TensorShape([])))

    dataset = dataset.map(tf_encode)
    dataset = dataset.filter(filter_sequence_length)
    # cache the dataset to memory to get a speedup while reading from it.
    dataset = dataset.cache()
    dataset = dataset.shuffle(BUFFER_SIZE).padded_batch(global_batch_size)
    dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)

    return dataset


if __name__ == "__main__":

    parser = create_argument_parser()

    parser.add_argument(
        '--max-sequence-length',
        type=int, required=False, default=60,
        help='Max. sequence length')

    parser.add_argument(
        '--num-attention-heads',
        type=int, required=False, default=12,
        help='Num. attention heads')

    parser.add_argument(
        '--feed-forward-layer-size',
        type=int, required=False, default=3072,
        help='Element-wise feed forward layer size')

    args = parser.parse_args()

    if args.remote_debug:
        import pydevd_pycharm
        pydevd_pycharm.settrace('192.168.178.85', port=57491, stdoutToServer=True, stderrToServer=True)

    mlp.logging.use_fancy_colors()

    logger_name = os.path.basename(__file__)
    logger = get_logger(logger_name)

    # TODO : seed
    # seed = args.seed
    # logger.info(f"Seed : {seed}")
    # np.random.seed(args.seed)

    use_mixed_precision = args.float16

    ##################################################
    #
    # [START] Setup
    #
    ##################################################

    # ############ Conversations dataset #############
    dataset_path = args.dataset_path
    base_dataset_filename = args.base_dataset_filename
    logger.info(f"dataset_path : {dataset_path}")
    logger.info(f"base_dataset_filename : {base_dataset_filename}")

    max_sequence_length = args.max_sequence_length
    ##################################################

    # ############ Model configuration ###############
    embedding_size = args.embedding_size
    logger.info(f"embedding_size : {embedding_size}")

    state_size = args.state_size
    logger.info(f"state_size : {state_size}")

    num_layers = args.num_layers
    logger.info(f"num_layers : {num_layers}")

    num_attention_heads = args.num_attention_heads
    logger.info(f"num_attention_heads : {num_attention_heads}")

    feed_forward_layer_size = args.feed_forward_layer_size

    dropout = args.dropout
    logger.info(f"dropout rate : {dropout}")
    ##################################################

    # ########### Training/optimization ##############
    experiment_name = args.experiment_name
    logger.info(f"experiment_name: {experiment_name}")

    num_gpus = args.num_gpus
    num_gpus_str = "all available" if num_gpus is None else num_gpus
    logger.info(f"Num. GPUs to use for training: {num_gpus_str}")

    batch_size_per_replica = args.batch_size
    logger.info(f"Batch size per replica: {batch_size_per_replica}")

    clip = args.gradient_clipping
    willClip = clip > 0.0
    logger.info(f"Gradient clipping : {clip} (Will clip? {willClip})")

    learning_rate = args.learning_rate
    logger.info(f"Base learning rate : {learning_rate}")

    decoder_learning_ratio = args.decoder_learning_rate_ratio
    logger.info(f"Decoder LR ratio : {learning_rate}")

    num_epochs = args.num_epochs
    logger.info(f"Number of training epochs : {num_epochs}")

    progress_logging_period = args.progress_logging_period
    logger.info(f"Progress logging period : {progress_logging_period}")

    # For Tensorboard
    metric_names = {
        'batch.loss': 'cross_entropy',
        'duration.batch': 'training_time',
        'duration.window_average': 'training_time',
        'duration.dataset': 'training_time',
        'batch_size': 'size'
    }

    # ########### Load datasets ##############
    logger.info('Setup data sets ...')

    logger.info('Loading training set ...')
    training_dataset, voc = load_sentence_pair_data(dataset_path_for('training'), logger)
    logger.info('Loading validation set ...')
    validation_dataset, _unused_ = load_sentence_pair_data(dataset_path_for('validation'), logger)

    # training_dataset = training_dataset[:40000]
    # validation_dataset = validation_dataset[:12000]

    logger.debug(f"Number of sentence pairs in training set: {len(training_dataset)}")
    logger.debug(f"Number of sentence pairs in validation set: {len(validation_dataset)}")

    all_training_sentences = []
    for pair in training_dataset:
        all_training_sentences += pair

    logger.debug(f"Building tokenizer vocabulary ...")
    tokenizer = tfds.deprecated.text.SubwordTextEncoder.build_from_corpus(
        all_training_sentences, target_vocab_size=2**13)

    vocab_size = tokenizer.vocab_size + 2

    del all_training_sentences
    logger.debug(f"Building tokenizer vocabulary ... READY")

    tf_encode = create_chatbot_tf_encode_func(tokenizer)
    filter_sequence_length = create_length_filter_func(max_sequence_length)
    ############################################

    # ########### Build model and optimizers ##############
    devices = None
    if num_gpus is not None:
        devices = [f"/gpu:{i}" for i in range(num_gpus)]

    strategy = tf.distribute.MirroredStrategy(devices=devices)
    with strategy.scope():
        logger.info('Building model ...')

        transformer = Transformer(num_layers, state_size,
                                  num_attention_heads, feed_forward_layer_size,
                                  vocab_size, vocab_size,
                                  pe_input=vocab_size,
                                  pe_target=vocab_size,
                                  rate=dropout)

        train_model = TrainModel(transformer)

        # Initialize optimizers
        # TODO : learning rate provided by arguments not used
        logger.info('Building optimizers ...')
        learning_rate = CustomSchedule(state_size)

        # learning_rate = 5e-3
        optimizer = tf.keras.optimizers.Adam(learning_rate, beta_1=0.9, beta_2=0.98, epsilon=1e-9)

    # ####################################################

    # ############### Prepare datasets ###################
    global_batch_size = batch_size_per_replica*strategy.num_replicas_in_sync
    logger.info(f"Global batch size : {global_batch_size}")

    training_dataset = prepare_dataset(training_dataset, tf_encode, filter_sequence_length)
    validation_dataset = prepare_dataset(validation_dataset, tf_encode, filter_sequence_length)

    len_training_dataset = training_dataset.reduce(0, lambda x, _: x + 1).numpy()
    len_validation_dataset = validation_dataset.reduce(0, lambda x, _: x + 1).numpy()

    logger.info(f"Number of training batches : {len_training_dataset}")
    logger.info(f"Number of validation batches : {len_validation_dataset}")

    # Distribute training and validation set
    training_dist_dataset = strategy.experimental_distribute_dataset(training_dataset)
    validation_dist_dataset = strategy.experimental_distribute_dataset(validation_dataset)
    # ####################################################

    # ############## SETUP TRAINING ##################
    logger.info('Prepare training ...')

    trainer = mlp.trainers.DefaultTrainer(optimizer,
                                          transformer,
                                          # use_mixed_precision=use_mixed_precision,
                                          distribution_strategy=strategy,
                                          batch_data_signature=training_dist_dataset.element_spec)

    average_loss_evaluator = mlp.evaluation.MetricEvaluator(trainer=trainer,
                                                            distribution_strategy=strategy,
                                                            batch_metric_funcs={
                                                                "loss": mlp.evaluation.GatherMaskedLoss(strategy)
                                                            },
                                                            name="AverageLossEvaluator")

    callbacks = [mlp.callbacks.TrainingMetricsLogger(metric_evaluator=average_loss_evaluator),
                 mlp.callbacks.TestMetricsLogger(validation_dist_dataset,
                                                 'validation',
                                                 metric_evaluator=average_loss_evaluator,
                                                 batch_averaging_window=len_validation_dataset),
                 # mlp.callbacks.BatchSizeLogger(),
                 mlp.callbacks.CheckpointManager(metric_to_monitor='validation.window_average.loss',
                                                 base_checkpoint_filename=args.experiment_name,
                                                 archive_last_model_checkpoint_every=20000),
                 mlp.callbacks.LogProgress(log_period=progress_logging_period, set_names=["training", "validation"]),
                 mlp.callbacks.AutoTensorboard(experiment_name=experiment_name, dataset_name='training',
                                               metric_names=metric_names),
                 mlp.callbacks.AutoTensorboard(experiment_name=experiment_name, dataset_name='validation',
                                               metric_names=metric_names),
                 # Batch-level batch duration and batch size
                 mlp.callbacks.Tensorboard(['duration.batch', 'batch_size'],
                                           experiment_name=experiment_name,
                                           dataset_name='training_params',
                                           metric_names=metric_names,
                                           ignore_missing_metrics=True),
                 # Batch-level average batch duration
                 mlp.callbacks.Tensorboard(['duration.window_average'],
                                           experiment_name=experiment_name,
                                           dataset_name='training_params',
                                           metrics_are_averages=True,
                                           metric_names=metric_names,
                                           ignore_missing_metrics=True),
                 # Epoch-level epoch duration
                 mlp.callbacks.Tensorboard(['duration.dataset'],
                                           experiment_name=experiment_name,
                                           dataset_name='training_params',
                                           batch_level=False,
                                           metric_names=metric_names,
                                           ignore_missing_metrics=True)]

    manager = mlp.trainers.TrainingManager(trainer,
                                           training_dist_dataset,
                                           num_batches_per_epoch=len_training_dataset,
                                           num_epochs=num_epochs,
                                           callbacks=callbacks,
                                           experiment_data=args)

    tc_file = args.training_checkpoint
    if tc_file:
        logger.info(f"Loading training checkpoint : {tc_file}")

        with open(tc_file, 'rb') as f:
            checkpoint = pickle.load(f)

        manager.set_state(checkpoint)

        logger.info(f"Ready loading training checkpoint.")

        del checkpoint

    trainer.set_training_model(train_model)

    ##################################################

    logger.info('Start training ...')
    manager.start_training()







